/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <math.h>
#include <map>

#include "Profiler.h"
#include "bytecodes.h"

#include <vm/Universe.h>

uint64_t bytecode_use_count[] = {
    0, // BC_HALT
    0, // BC_DUP
    0, // BC_PUSH_LOCAL
    0, // BC_PUSH_ARGUMENT
    0, // BC_PUSH_FIELD
    0, // BC_PUSH_BLOCK
    0, // BC_PUSH_CONSTANT
    0, // BC_PUSH_GLOBAL
    0, // BC_POP
    0, // BC_POP_LOCAL
    0, // BC_POP_ARGUMENT
    0, // BC_POP_FIELD
    0, // BC_SEND
    0, // BC_SUPER_SEND
    0, // BC_RETURN_LOCAL
    0  // BC_RETURN_NON_LOCAL
};

struct ltstr {
    bool operator()(const char* s1, const char* s2) const {
        return strcmp(s1, s2) < 0;
    }
};

struct ltsize {
    bool operator()(const size_t s1, const size_t s2) const {
        return s1 < s2;
    }
};

// called methods, primitives
std::map<const char*, uint64_t, struct ltstr> called_methods;

// object sizes
std::map<const size_t, uint64_t, struct ltsize> allocated_objects;

//file to write the xml profile to
FILE* fp;

//flag if the profiler is initialized
int initialized = 0;

/* Resets the profiler. Automatically called on first access to profiler if it
   was not done explicitly before.   */
void Profiler_initialize() {
	initialized = 1;
}

/* Adds a bytecode to the profile. */
void Profiler_add(uint8_t bc) {
    bytecode_use_count[bc]++;
}

void Profiler_add_primitive(char* method_sig) {
    char* key = strdup(method_sig);
    called_methods[key] = 0;
}

void Profiler_method_called(const char* method_sig) {
    called_methods[method_sig]++;
}

void Profiler_malloc(size_t size) {
    allocated_objects[size]++;
}

/* Closes the profile, by closing the xml output file, freeing the allocated
   stats memory and marking the profiler as uninitialized        */
void Profiler_close() {
	/*if (fp != NULL) {
		fprintf(fp, "\t</basic_block>\n</trace>\n");
		fclose(fp);
	}*/
	initialized = 0;
}

void Profiler_print_top_n(int top_n, int seq_length) {

}

/*prints top twenty combinations of doubles, triples and quads*/
void Profiler_print_top(int top_n) {
	Profiler_print_top_n(top_n, 2);
	Profiler_print_top_n(top_n, 3);
	Profiler_print_top_n(top_n, 4);
}


/* Prints the complete preprocessed profile as a table into a html document */
void Profiler_print_html(char* filename) {
	FILE *fp;

	if((fp=fopen(filename, "wb"))==NULL) {
		 printf("Can't open file %s to write.\n", filename);
         Universe_exit(1);
	}

	fprintf(fp ,"<html><head></head><body>\n");
	fprintf(fp ,"<table>\n");
	fprintf(fp ,"\t<tr>\n");
	fprintf(fp ,"\t\t<th>Bytecode</th>\n");
    fprintf(fp ,"\t\t<th>Count</th>\n");
    fprintf(fp ,"\t</tr>\n");
    
	for (size_t i = 0; i < 16; i++)	{
		fprintf(fp ,"\t\t<tr><td>%s</td><td>%d</td></tr>\n", bytecode_names[i], bytecode_use_count[i]);
	}
    
	fprintf(fp ,"</table>\n");
    
    fprintf(fp ,"<table>\n");
	fprintf(fp ,"\t<tr>\n");
	fprintf(fp ,"\t\t<th>Signature</th>\n");
    fprintf(fp ,"\t\t<th>Count</th>\n");
    fprintf(fp ,"\t</tr>\n");

    for(std::map<const char*, uint64_t, struct ltstr>::const_iterator it = called_methods.begin(); it != called_methods.end(); ++it) {
        fprintf(fp ,"\t\t<tr><td>%s</td><td>%d</td></tr>\n", it->first, it->second);
    }
    
    fprintf(fp ,"</table>\n");

    fprintf(fp ,"<table>\n");
	fprintf(fp ,"\t<tr>\n");
	fprintf(fp ,"\t\t<th>Object Size</th>\n");
    fprintf(fp ,"\t\t<th>Count</th>\n");
    fprintf(fp ,"\t</tr>\n");
    
    for(std::map<const size_t, uint64_t, struct ltsize>::const_iterator it = allocated_objects.begin(); it != allocated_objects.end(); ++it) {
        fprintf(fp ,"\t\t<tr><td>%d</td><td>%d</td></tr>\n", it->first, it->second);
    }
    
    fprintf(fp ,"</table>\n");
	fprintf(fp ,"\n</body></html>\n");

	fclose(fp);
}


/* Prints the complete preprocessed profile as a table into a html document */
void Profiler_print_txt(char* filename) {
	FILE *fp;
    
	if((fp=fopen(filename, "wb"))==NULL) {
        printf("Can't open file %s to write.\n", filename);
        Universe_exit(1);
	}
    
	fprintf(fp, "Bytecode:\t");
	for (size_t i = 0; i < 16; i++)	{
		fprintf(fp, "%s\t", bytecode_names[i]);
	}
    
    fprintf(fp, "\nCount:\t");
    for (size_t i = 0; i < 16; i++)	{
		fprintf(fp, "%d\t", bytecode_use_count[i]);
	}
    
    fprintf(fp, "\n\n");
    
    fprintf(fp, "Signature\t");
    for(std::map<const char*, uint64_t, struct ltstr>::const_iterator it = called_methods.begin(); it != called_methods.end(); ++it) {
        fprintf(fp, "%s\t", it->first);
    }

    fprintf(fp ,"\nCount:\t");
    for(std::map<const char*, uint64_t, struct ltstr>::const_iterator it = called_methods.begin(); it != called_methods.end(); ++it) {
        fprintf(fp, "%d\t", it->second);
    }
    
    fprintf(fp, "\n\n");

	fprintf(fp, "ObjectSize:\t");

    for(std::map<const size_t, uint64_t, struct ltsize>::const_iterator it = allocated_objects.begin(); it != allocated_objects.end(); ++it) {
        fprintf(fp, "%d\t", it->first);
    }
    
    fprintf(fp, "\nCount:\t");
    for(std::map<const size_t, uint64_t, struct ltsize>::const_iterator it = allocated_objects.begin(); it != allocated_objects.end(); ++it) {
        fprintf(fp ,"%d\t", it->second);
    }
    
    fprintf(fp, "\n\n");
    
	fclose(fp);
}

void Profiler_print() {

	
}
