grammar AspectC;

options {
	output = AST;
  language = Java;
}

tokens {
    ADVICE		= 'advice';
    ASPECT		= 'aspect';
    POINTCUT	= 'pointcut';
    STATIC    = 'static';
    EXTERN    = 'extern';
    CONST     = 'const';
    HEXPREFIX = '0x';
    RESTRICT  = 'restrict';
}

@header {
  package de.sm.aspectc;
  
  import java.io.File;
}

@lexer::header {
  package de.sm.aspectc;
}

@members {
	private File parsedFile;
  
  /**
  * Make like error reporting for parser errors allows to benefit
  * from IDE make integration.
  */
  @Override
  public void reportError(RecognitionException re) {
    String errorMsg = "";
    if (parsedFile != null) {
      errorMsg += parsedFile.getAbsolutePath();
    }
    
    errorMsg += ":" + re.line + ": warning: Non critical parse error, file is ignored: ";
    System.err.print(errorMsg);
    super.reportError(re);
  }
  
  /**
  * Set the file which is actually parsed.
  * Is used to be able to generate meaningful error messages.
  * Is not used for parsing at all.
  */
  public void setParsedFile(File file) {
    parsedFile = file;
  }

	private int adviceCount = 0;
	private int pointcutCount = 0;
  
  public int getAdviceCount() {
    return adviceCount;
  }
  
  public int getPointcutCount() {
    return pointcutCount;
  }
}

aspectDefinition:
  ( preprocessor | aspect | NEWLINE )* EOF
;

aspect:
	ASPECT IDENTIFIER LBRACE 
	( advice | pointcut | NEWLINE )* 
	RBRACE SEMICOLON
;

advice:
	ADVICE adviceHeader content
	{ adviceCount++; }
;

adviceHeader :
	 ( ~(LBRACE | RBRACE | SEMICOLON) )+
;

preprocessor :
  HASH
  ((~( PPCONTINUE | NEWLINE ))* PPCONTINUE)*
  (~( PPCONTINUE | NEWLINE ))*
  NEWLINE
;

contentcontent:
	(content | (~(LBRACE | RBRACE))+)*;

content returns [String content]: 
  LBRACE
	contentcontent
	{ $content = $contentcontent.text; }
	RBRACE
;

pointcut :
	POINTCUT 
	(~(SEMICOLON))*
	SEMICOLON NEWLINE*
	{ pointcutCount++; } 
;



IDENTIFIER : LETTER ( DIGIT | LETTER | '-' | '_' )*;

DIGIT
    :    '0'..'9'
    ;
fragment LETTER
    : 'a'..'z'
    | 'A'..'Z'
    ;

LPARENTHESIS: '(';
RPARENTHESIS: ')';

SPECIAL : ':' | '='| '.' | '*' | '+' | '%' | '/' | '\\'
	| '_' | ',' | '|' | '&' | '-' | '"' | '<' | '>' | '[' | ']' | '!' | '\'' | '?' ;

PPCONTINUE: '\\' NEWLINE;

NEWLINE
  : '\r'? '\n'
  //{ channel=HIDDEN; }
;

WS
  : (' '|'\t')+		//|'\n'|'\r' 
  { channel=HIDDEN; }
;

COMMENT
  : '//' (~('\n'|'\r'))* NEWLINE
  { channel=HIDDEN; }
;
  
// multiple-line comments
ML_COMMENT
  : '/*' (options {greedy=false;} : .)* '*/' 
  { channel=HIDDEN; }
;

HASH : '#';
SEMICOLON : ';';
LBRACE : '{';
RBRACE : '}';
