package de.sm.metrics;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.Lexer;
import org.antlr.runtime.Token;
import org.omg.CORBA.portable.ApplicationException;

import de.sm.aspectc.AspectCLexer;
import de.sm.csimple.CSimpleLexer;

/**
 * This class implements basic Lines of code metrics.
 * @author Stefan Marr
 */
public class LinesOfCode extends Metrics {
	
	public LinesOfCode(File file) {
		super(file);
	}

	/**
	 * Count the total number of lines in the file.
	 * @return
	 * @throws IOException
	 */
	public final int getLinesInFile() throws IOException {
		FileReader fr = new FileReader(file);
		LineNumberReader lnr = new LineNumberReader(fr, 1);
		
		while (lnr.readLine() != null) { /* NOP */ }
		
		return lnr.getLineNumber() + 1;
	}
	
	private Lexer lexer;
	private int eofConst;
	private int newlineConst;
	
	/**
	 * Count the number of non-white spaces lines.
	 * This are the line contributing to the functionality without comments. 
	 * @return
	 * @throws IOException 
	 */
	public final int getLineCountWithStatements() throws IOException {
		ANTLRFileStream input = new ANTLRFileStream(file.getAbsolutePath());
		initForFileType(input);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		
		int previousLine = -1;
		int lineCount = 0;
		Token token = tokens.LT(1);
		
		while (token.getType() != eofConst) {
			if (token.getType() != newlineConst
					&& previousLine != token.getLine()) {
				//System.out.println(token.getLine() + " " + token.getText());
				lineCount++;
				previousLine = token.getLine();
			}
			
			tokens.consume();
			token = tokens.LT(1);
		}
		
		return lineCount;
	}
	
	private void initForFileType(ANTLRFileStream input) {
		if (file.getName().endsWith(".c") || file.getName().endsWith(".h")) {
			lexer = new CSimpleLexer(input);
			eofConst = CSimpleLexer.EOF;
			newlineConst = CSimpleLexer.NEWLINE;
		} else if (file.getName().endsWith(".ah")) {
			lexer = new AspectCLexer(input);
			eofConst = AspectCLexer.EOF;
			newlineConst = AspectCLexer.NEWLINE;
		} else if (file.getName().endsWith(".vmadl")) {
			lexer = new AspectCLexer(input);
			eofConst = AspectCLexer.EOF;
			newlineConst = AspectCLexer.NEWLINE;
		} else {
			throw new IllegalArgumentException(
					"Unsupported filetype: " + file.getName());
		}
		
	}
}
