grammar Vmadl;

options {
	output = AST;
  language = Java;
  //backtrack = true;
}

tokens {
    SERVICE		= 'service';
    STARTUP		= 'startup';
    SHUTDOWN	= 'shutdown';
    EXPOSE		= 'expose';
    ADVICE		= 'advice';
    POINTCUT	= 'pointcut';
    REPLACE		= 'replace';
    REQUIRE		= 'require';
    BOOT		  = 'boot'; // would be nice if i could get rid of this
    FOR				= 'for';
    TYPES			= 'types';
    COMBINE   = 'combine';
    STATIC    = 'static';
    EXTERN    = 'extern';
    CONST     = 'const';
    HEXPREFIX = '0x';
    RESTRICT  = 'restrict';
}

@header {
  package de.sm.vmadl;
  
  import de.sm.vmadl.elements.*;
  import java.io.File;
  import de.sm.odl.elements.TypeRef;
}

@lexer::header {
  package de.sm.vmadl;
}

@members {
	private ServiceRegistry reg;
	private CombinationRegistry combReg;
	private File parsedFile;
	
	/**
	* Make like error reporting for parser errors allows to benefit
	* from IDE make integration.
	*/
	@Override
	public void reportError(RecognitionException re) {
	  String errorMsg = "";
	  if (parsedFile != null) {
	    errorMsg += parsedFile.getAbsolutePath();
	  }
	  
	  errorMsg += ":" + re.line + ": error: ";
	  System.err.print(errorMsg);
	  super.reportError(re);
	}
	
	/**
	* Set the file which is actually parsed.
	* Is used to be able to generate meaningful error messages.
	* Is not used for parsing at all.
	*/
	public void setParsedFile(File file) {
	  parsedFile = file;
	}
}

architecture [ServiceRegistry reg, CombinationRegistry combReg]
 returns [ List<ServiceModule> modules, List<Combination> combinations ] :
{ 
	$modules = new ArrayList<ServiceModule>();
	$combinations = new ArrayList<Combination>();
	
	this.reg = $reg;
	this.combReg = $combReg;
}
    /* Include only service nodes into tree, additional declarations are handled using raw data */
    ( 
     NEWLINE!
     | service { $modules.add($service.s);}
     | combination { $combinations.add($combination.c); }
     )* EOF!
;

interaction:
  ( NEWLINE! | combination )* EOF!
;

combination returns [Combination c]:
  COMBINE a=IDENTIFIER ',' b=IDENTIFIER 
  { $c = combReg.get(new CombinationId(reg.get($a.text), reg.get($b.text))); }
  LBRACE ( combinationContent[$c] )* RBRACE
  { $c.lazyInitWith(retval, parsedFile); }
;


service returns [ServiceModule s] :
 	SERVICE IDENTIFIER { $s = reg.get($IDENTIFIER.text); }
  LBRACE ( basicModuleContent[$s] )* RBRACE
  { $s.initialized = true; }
;


basicModuleContent [ ServiceInterface s ]:
		 require				{ $s.requirements.add($require.service); }
  	 | expose				{ $s.pointcuts = $expose.pointcuts; }
  	 | declarations { $s.declarations.add($declarations.text); }
  	 //| optionalRelations { $s.relations.add($optionalRelations.s); }
  	 | types				{ $s.typesSection = $types.content; }
  	 | namedPart[$s]
  	 | replace[$s]
  	 | NEWLINE
;

combinationContent [Combination c]:
  require           { $c.requirements.add($require.service); }
  | advice          { $c.advices.add($advice.adv); }
  | startup
  | shutdown
  | preprocessor    { $c.preprocessor.add($preprocessor.text); }
  | variable        { $c.variables.add($variable.var); }
  | function        { $c.functions.add($function.func); }
  | NEWLINE
;

literalValue:
  IDENTIFIER | DIGIT+ | hexValue
;

variable returns [ Variable var ]:
  { $var = new Variable(); }
  (STATIC { $var.isStatic = true; }| EXTERN { $var.isExtern = true; })?
  typeRef     { $var.type = $typeRef.ref; } 
  IDENTIFIER  { $var.name = $IDENTIFIER.text; }
  ('[' arrDim=literalValue ']' { $var.arrayDim = $arrDim.text; })?
  ( '=' val=literalValue { $var.value = $val.text; })?
  ';'
;

typeRef returns [ TypeRef ref ]:
  { $ref = new TypeRef(); }
  (CONST { $ref.constQ = true; })?
  IDENTIFIER { $ref.name = $IDENTIFIER.text; }
  ('*' { $ref.incPtr(); } )*
  (RESTRICT { $ref.restrictQ = true; })?
  (CONST { $ref.constQ2 = true; })?
;

hexValue:
  HEXPREFIX (DIGIT | 'a'..'f' | 'A'..'F')+
;

function returns [ Function func ]:
  { $func = new Function(retval, parsedFile); }
  (STATIC { $func.isStatic = true; }| EXTERN {$func.isExtern = true; })?
  returnType=typeRef name=IDENTIFIER
  { $func.name = $name.text;
    $func.returnType = $returnType.ref;
  } 
  LPARENTHESIS
    (paramType=typeRef paramName=IDENTIFIER?  {$func.addParam(($paramName != null)?$paramName.text:null, $paramType.ref);} 
      (',' moreParam=typeRef moreName=IDENTIFIER? {$func.addParam(($moreName != null)?$moreName.text:null, $moreParam.ref);}
      )*
    )? 
  RPARENTHESIS
  (content { $func.body = $content.content; }| ';' { $func.isDecl = true; })
;

namedPart [ServiceInterface s] returns [ ServiceInterface m ]:
  IDENTIFIER { QualifiedIdentifier qid = new QualifiedIdentifier(((ServiceModule)$s).name, $IDENTIFIER.text); 
  						 $m = new ServiceInterface(); 
  						 $m.partName = qid;
  						 $s.parts.put(qid, $m); 
  			     } 
	LBRACE ( basicModuleContent[$m] )* RBRACE
;

replace [ ServiceInterface s ] returns [ ServiceInterface m ]:
	REPLACE id=qualifiedIdentifier
	{ $m = new ServiceInterface(); 
	  $m.partName = $id.QId;
	  $s.replacements.put($id.QId, $m); 
  } 
	LBRACE ( basicModuleContent[$m] )* RBRACE
;

advice returns [Advice adv]:
	ADVICE adviceHeader content
  { $adv = new Advice(retval, parsedFile);
    $adv.header = $adviceHeader.text;
    $adv.content = $content.content;
  }
;

adviceHeader :
	 ( ~(LBRACE | RBRACE | ';') )+
;

declarations :
	preprocessor | declaration
;

preprocessor :
  '#'
  ((~( PPCONTINUE | NEWLINE ))* PPCONTINUE)*
  (~( PPCONTINUE | NEWLINE ))*
  NEWLINE
;

declaration :
	(~('#' | NEWLINE | LBRACE | RBRACE | ';' | SERVICE | STARTUP | SHUTDOWN
	   | EXPOSE | ADVICE | POINTCUT | REQUIRE | BOOT | FOR | REPLACE ))
	
	( ~(LBRACE | RBRACE | ';') )+
	(content ';'? | ';' )
;

contentcontent:
	(content | (~(LBRACE | RBRACE))+)*;

content returns [String content]: 
  LBRACE
	contentcontent
	{ $content = $contentcontent.text; }
	RBRACE
;

types returns [String content]:
	TYPES c=content { $content = $c.content; };

startup
  : STARTUP
	(LBRACE (expose | advice | boot )* RBRACE);

shutdown
  : SHUTDOWN
	(LBRACE (expose | advice)* RBRACE);	
	
boot :
	BOOT
	( ~(LBRACE | RBRACE | ';') )+
	';'
;
	
expose returns [List<String> pointcuts] :
{ $pointcuts = new ArrayList<String>(); }
  EXPOSE NEWLINE* LBRACE NEWLINE* 
	( pointcut {$pointcuts.add($pointcut.text); }
	)*
	RBRACE
;

require returns [ ServiceModule service ] : 
	REQUIRE IDENTIFIER { $service = reg.get($IDENTIFIER.text); } 
	';';
	
pointcut :
	POINTCUT 
	(~(';'))*
	';' NEWLINE* 
;

/*optionalRelations returns [OptionalRelation s] :
	FOR IDENTIFIER { $s = new OptionalRelation(); 
										 $s.service = reg.get($IDENTIFIER.text); }
	LBRACE ( basicModuleContent[$s] )* RBRACE
;*/

qualifiedIdentifier returns [ QualifiedIdentifier QId ] : 
	module=IDENTIFIER '.' name=IDENTIFIER
	{ $QId = new QualifiedIdentifier($module.text, $name.text); }
; 

IDENTIFIER : LETTER ( DIGIT | LETTER | '-' | '_' )*;

DIGIT
    :    '0'..'9'
    ;
fragment LETTER
    : 'a'..'z'
    | 'A'..'Z'
    ;

LPARENTHESIS: '(';
RPARENTHESIS: ')';

SPECIAL : ':' | '='| '.' | '*' | '+' | '%' | '#' | '/' | '\\'
	| '_' | ';' | ',' | '|' | '&' | '-' | '"' | '<' | '>' | '[' | ']' | '!' | '\'' | '?' ;

PPCONTINUE: '\\' NEWLINE;

NEWLINE
  : '\r'? '\n'
  //{ channel=HIDDEN; }
;

WS
  : (' '|'\t')+		//|'\n'|'\r' 
  { channel=HIDDEN; }
;

COMMENT
  : '//' (~('\n'|'\r'))* NEWLINE
  { channel=HIDDEN; }
;
  
// multiple-line comments
ML_COMMENT
  : '/*' (options {greedy=false;} : .)* '*/' 
  { channel=HIDDEN; }
;

//SEMICOLON : ';';
LBRACE : '{';
RBRACE : '}';
