package de.sm.vmadl;

import java.io.File;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Collection;
import java.util.Formatter;
import java.util.HashSet;
import java.util.Set;
import java.util.Map.Entry;
import java.util.logging.ConsoleHandler;
import java.util.logging.Logger;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;

import de.sm.misc.MakeFormatter;
import de.sm.odl.Odl;
import de.sm.odl.OdlLexer;
import de.sm.odl.OdlParser;
import de.sm.vmadl.elements.Function;
import de.sm.odl.elements.FunctionParam;
import de.sm.vmadl.VmadlParser.architecture_return;
import de.sm.vmadl.elements.Advice;
import de.sm.vmadl.elements.ArchitectureElement;
import de.sm.vmadl.elements.Combination;
import de.sm.vmadl.elements.LanguageElement;
import de.sm.vmadl.elements.QualifiedIdentifier;
import de.sm.vmadl.elements.ServiceInterface;
import de.sm.vmadl.elements.ServiceModule;
import de.sm.vmadl.elements.Variable;

/**
 * Implements the compilation of Vmadl to the actual implementation
 * language files.
 *
 * @author Stefan Marr
 */
public class Vmadl {
	
	/**
	 * Registry of all referenced service modules.
	 */
	private ServiceRegistry reg;
	
	/**
	 * Registry for module combination definitions.
	 */
	private CombinationRegistry combReg;
	
	/**
	 * Template format string for the header file.
	 */
	private String headerFile;
	
	/**
	 * Template format string for the aspect file.
	 */
	private String aspectFile;
	
	/**
	 * Source directory.
	 */
	private File sourceDir;
	
	/**
	 * Output directory used for generated files.
	 */
	private File outputBase;
	
	/**
	 * Logger for this class.
	 */
	private Logger log;
	
	/**
	 * ODL compiler used for ODL type definitions.
	 */
	private Odl odl;
	
	/**
	 * Service modules chosen for current configuration.
	 */
	private Collection<ServiceModule> serviceModules;
	
	/**
	 * Initializes all important registries and templates.
	 */
	public Vmadl(final ServiceRegistry serviceReg,
			final CombinationRegistry combReg, final Odl odl) {
		log = Logger.getLogger("VMADL");
		ConsoleHandler ch = new ConsoleHandler();
		ch.setFormatter(new MakeFormatter());
		log.addHandler(ch);
		reg = serviceReg;
		this.odl = odl;
		this.combReg = combReg;
		
		initTemplates();
	}
	
	/**
	 * Initialize the used templates.
	 */
	private void initTemplates() {
		// %1 - Include Guard
		// %2 - Name of the vmadl file
		// %3 - Content
		headerFile = "#ifndef %1$s\n#define %1$s\n/**\n * This files is "
			+ "generated, do not modify, edit the VMADL file directly instead."
			+ "\n * Generated from: %2$s\n"
			+ " */\n\n"
			+ "%3$s\n"
			+ "\n#endif /* %1$s */\n";
		
		// %1 - Include guard
		// %2 - Name of the vmadl file
		// %3 - Includes
		// %4 - Aspect name
		// %5 - Aspect content
		aspectFile = "#ifndef %1$s\n#define %1$s\n/**\n * This files is "
			+ "generated, do not modify, edit the VMADL file directly instead."
			+ "\n * Generated from: %2$s\n"
			+ " */\n\n"
			+ "\n%3$s\n\n"
			+ "aspect %4$s {\n"
			+ "%5$s\n\n"
			+ "};\n"
			+ "\n#endif /* %1$s */\n";
	}
	
	/**
	 * Write the given content to a C header file.
	 * 
	 * @param targetFile to write to
	 * @param includeGuard for C header file
	 * @param vmadlFilename where the implementation is extracted from
	 * @param content as string
	 */
	private void writeHeaderFile(final File targetFile, 
			final String includeGuard, final String vmadlFilename, 
			final String content) {
		try {
			Formatter formatter = new Formatter(targetFile);
			formatter.format(headerFile, includeGuard, vmadlFilename, content);
			formatter.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private void writeAspectFile(final File targetFile,
			final String includeGuard, final String vmadlFilename,
			final String includes, final String aspectName,
			final String content) {
		writeFile(targetFile, String.format(aspectFile, includeGuard, 
				vmadlFilename, includes, aspectName, content));
	}
	
	private void writeFile(final File targetFile, final String content) {
		FileWriter fw;
		try {
			fw = new FileWriter(targetFile);
			fw.write(content);
			fw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Loads the VM Architecture description from the given file.
	 * @param file to be parsed
	 */
	protected final void loadFromFile(final File file) {
		log.info("Load " + file.getAbsolutePath());
		CharStream input;
		try {
			input = new ANTLRFileStream(file.getAbsolutePath());
			VmadlLexer lexer = new VmadlLexer(input);
			
	        CommonTokenStream tokens = new  CommonTokenStream(lexer);
			VmadlParser parser = new VmadlParser(tokens);
			parser.setParsedFile(file);
		    
		    architecture_return arch = parser.architecture(reg, combReg);
		    
		    // Set file
		    for (ServiceModule module : arch.modules) {
		    	module.file = file;
		    }
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (RecognitionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void parseTypesSection(final ServiceModule module) {
		log.info("Parse types section of " + module.name);
		CharStream input = new ANTLRStringStream(module.typesSection);
		OdlLexer odlLexer = new OdlLexer(input);
		
		CommonTokenStream odlTokens = new CommonTokenStream(odlLexer);
		OdlParser odlParser = new OdlParser(odlTokens);
		
		try {
			odlParser.definitions(odl.getTypeRegistry(), module);
		} catch (RecognitionException e) {
			// TODO Auto-generated catch block
			System.err.println("RecognitionException while parsing types section of " + module.name);
			e.printStackTrace();
		}
	}
	
	public final void checkConsistency() throws CompilerException {
		// verify internal consistency of VMADL files
		
		// all used pointcuts in advices are only allowed to use function names
		// or pointcuts defined in other vmadl files, and therefore the right
		// require statements have to be given. Same could be proven for 
		// startup and shutdown, advice are only allowed to use pointcuts 
		// available at this time
		
		// check whether all requirements expressed by require are fulfilled
		for (ServiceModule module : serviceModules) {
			if (!module.initialized) {
				throw new CompilerException("Module " + module.name.getModule()
						+ " was chosen as part for the configuration, "
						+ "but had not been found in source tree."
						+ " Compilation failed.", module);
			}
			
			for (ServiceModule reqedModule : module.requirements) {
				if (!reqedModule.initialized) {
					throw new CompilerException("For module "
							+ module.name.getModule() + " the required module "
							+ reqedModule.name.getModule()
							+ " was not found. Compilation failed.", module);
				}
			}
		}

		// for combinations there should be everything alright,
		// because serviceModule has been used to get active combinations
	}
	
	public final void generateImplementation(final String outputDir) {
		outputBase = new File(outputDir);
		if (!outputBase.exists() && !outputBase.mkdirs()) {
			throw new RuntimeException("Output directory ("
					+ outputDir + ") could not be created.");
		}
		
		// There should not be any optional parts any more, TODO drop code later
		// At first resolve optional stuff (OptionalRelation)
		// resolveOptionalRelations();
		
		// prepare ODL
		initializeOdl();
		
		Set<ServiceModule> modulesWithDep = new HashSet<ServiceModule>();
		
		for (ServiceModule module : serviceModules) {
			if (module.initialized) {
				generateHeader(module);
				if (hasPointcuts(module)) { generateAspect(module); }
				if (module.types.size() > 0) { 
					generateTypes(module);
					modulesWithDep.add(module);
					modulesWithDep.addAll(module.requirements);
				}
			}
		}
		
		StringBuilder includes = new StringBuilder();
		for (ServiceModule module : modulesWithDep) {
			includes.append("#include <module.");
			includes.append(module.name);
			includes.append(">\n");
		}
		
		odl.generateImplementation(new File(outputDir, "odl.c"),
				includes.toString());
		
		// generate combination aspects
		Collection<Combination> activeCombs =
			combReg.getActiveCombinations(serviceModules);
		for (Combination comb : activeCombs) {
			generateCombinationImplementation(comb);
		}
	}
	
	// There should not be any optional parts any more, drop code later
	/*private void resolveOptionalRelations() {
		Collection<ServiceModule> services = reg.getServices();
		for (ServiceModule module : services) {
			for (OptionalRelation relation : module.relations) {
				if (relation.service.initialized) {
					module.advices.addAll(relation.advices);
					module.declarations.addAll(relation.declarations);
					module.pointcuts.addAll(relation.pointcuts);
					module.shutdownAdvices.addAll(relation.shutdownAdvices);
					module.shutdownPointcuts.addAll(relation.shutdownPointcuts);
					module.startupAdvices.addAll(relation.startupAdvices);
					module.startupPointcuts.addAll(relation.startupPointcuts);
					module.requirements.addAll(relation.requirements);
					module.requirements.add(relation.service);
				}
			}
		}
	}*/

	private void generateCombinationImplementation(Combination comb) {
		StringBuilder content = new StringBuilder();
		
		if (hasImplementation(comb)) {
			generateCombinationHeader(comb);
			generateCombinationImpl(comb);
		}	
		
		generateCombinationAspect(comb);
	}

	private void generateCombinationImpl(final Combination comb) {
		StringBuilder content = new StringBuilder();
		
		content.append("#include <");
		content.append("combine."
				+ comb.id.getModuleA().name + "-" + comb.id.getModuleB().name
				+ ".h");
		content.append(">\n\n");
		
		for (Variable var : comb.variables) {
			if (!var.isExtern) {
				if (var.isStatic) {
					content.append("static ");
				}
				content.append(var.type.getCompleteName());
				content.append(" ");
				content.append(var.name);
				if (var.arrayDim != null) {
					content.append(var.arrayDim);
				}
				if (var.value != null) {
					content.append(" = ");
					content.append(var.value);
				}
				content.append(";\n");
			}
		}
		
		for (Function func : comb.functions) {
			content.append("\n");
			renderFunctionHeader(content, func);
			content.append("{\n");
			
			renderDebugHint(content, func);
			content.append(func.body);
			content.append("}\n\n");
		}
		
		content.append("\n");
		
		File outputFile = new File(outputBase, "combine."
				+ comb.id.getModuleA().name + "-" + comb.id.getModuleB().name
				+ ".c");
		
		writeFile(outputFile, content.toString());
	}

	private void renderParameterList(StringBuilder content, final Function func) {
		boolean first = true;
		for (FunctionParam param : func.getParams()) {
			if (!first) {
				content.append(", ");
			} else {
				first = false;
			}
			content.append(param.type.getCompleteName());
			if (param.name != null) {
				content.append(" ");
				content.append(param.name);
			}
		}
	}
	
	/**
	 * Generate the header file for all implmenentations done in
	 * given combine definition.
	 * Forward decl for variables and functions are generated.
	 * @param comb
	 */
	private void generateCombinationHeader(final Combination comb) {
		StringBuilder content = new StringBuilder();
		
		// Self include (of the own header file)
		content.append("#include <module.");
		content.append(comb.id.getModuleA().name);
		content.append(">\n");
		
		// Self include (of the own header file)
		content.append("#include <module.");
		content.append(comb.id.getModuleB().name);
		content.append(">\n");
		
		for (String pre : comb.preprocessor) {
			content.append(pre);
		}
		
		for (Variable var : comb.variables) {
			content.append("extern ");
			content.append(var.type.getCompleteName());
			content.append(" ");
			content.append(var.name);
			content.append(";\n");
		}
		
		for (Function func : comb.functions) {
			renderFunctionHeader(content, func);
			content.append(";\n");
		}
		
		ServiceModule a = comb.id.getModuleA();
		ServiceModule b = comb.id.getModuleB();
		
		File outputFile = new File(outputBase, "combine." + a.name + "-"
				+ b.name + ".h");
		String guard = getGuardFromName(
				a.name.getModule() + b.name.getModule(), "h");
		writeHeaderFile(outputFile, guard, comb.getFile().getAbsolutePath(),
				content.toString());
	}

	private void renderFunctionHeader(StringBuilder content, Function func) {
		content.append(func.returnType.getCompleteName());
		content.append(" ");
		content.append(func.name);
		content.append("(");
		renderParameterList(content, func);
		content.append(")");
	}

	/**
	 * Estimate whether given combination needs an implementation
	 * file.
	 * @param comb
	 * @return
	 */
	private boolean hasImplementation(final Combination comb) {
		return !comb.functions.isEmpty() && !comb.variables.isEmpty();
	}

	private boolean hasPointcuts(final ServiceInterface module) {
		return !module.pointcuts.isEmpty()
			|| !module.shutdownPointcuts.isEmpty()
			|| !module.startupPointcuts.isEmpty();
	}
	
	private void generateTypes(final ServiceModule module) {
		StringBuilder includes = new StringBuilder();
		odl.generateForwardHeader(new File(outputBase,
				"module." + module.name + ".fwd.h"), module);

		includes = new StringBuilder();
		generateIncludes(module, includes);
		includes.append("#include <module.");
		includes.append(module.name);
		includes.append(">\n");
		odl.generateDefinitionHeader(new File(outputBase,
				"module." + module.name + ".def.h"), module,
				includes.toString());
	}
	
	private void generateCombinationAspect(final Combination comb) {
		StringBuilder content = new StringBuilder();
		StringBuilder includes = new StringBuilder();
		
		content.append("\t/*** Advices ***/\n");
		for (Advice advice : comb.advices) {
			renderAdvice(content, advice);
		}
		
		content.append("\t/*** Advices only active at startup time ***/\n");
		content.append("\t/* REMARK: propably there should be a guard "
				+ "to ensure this, automatically generated. */");
		for (Advice advice : comb.startupAdvices) {
			renderAdvice(content, advice);
		}
		
		content.append("\t/*** Advices only active at shutdown time ***/\n");
		content.append("\t/* REMARK: propably there should be a guard "
				+ "to ensure this, automatically generated. */");
		for (Advice advice : comb.shutdownAdvices) {
			renderAdvice(content, advice);
		}
			
		ServiceModule a = comb.id.getModuleA();
		ServiceModule b = comb.id.getModuleB();
		
		generateAdviceIncludes(includes, a, b);
		renderRequires(comb, includes);
		
		if (hasImplementation(comb)) {
			includes.append("#include <combine.");
			includes.append(a.name);
			includes.append("-");
			includes.append(b.name);
			includes.append(".h>\n");
		}
		
		File outputFile = new File(outputBase, "combine." + a.name + "-"
				+ b.name + ".ah");
		String guard = getGuardFromName(
				a.name.getModule() + b.name.getModule(), "ah");
		writeAspectFile(outputFile, guard, null, includes.toString(),
				a.name.getModule() + b.name.getModule(), content.toString());
	}

	/**
	 * Adds a line definition of the source file to the content of
	 * the advice to enable eclipse debugging capabilities in place.
	 * @param content
	 * @param advice
	 */
	private void renderAdvice(final StringBuilder content, final Advice advice) {
		content.append("advice ");
		content.append(advice.header);
		content.append("{\n");
		renderDebugHint(content, advice);
		content.append(advice.content);
		content.append("}\n\n");
	}

	private void renderDebugHint(final StringBuilder content,
			final LanguageElement element) {
		// TODO: add config param to enable this, but for now it is bad for debugging stuff, since the AspectC++ stuff changes var names and makes it imposible for debugging
		/*content.append("#line ");
		content.append(element.getStart().getLine());
		content.append(" \"");
		content.append(element.getFile().getAbsolutePath());
		content.append("\"\n");*/
	}

	private void generateAdviceIncludes(final StringBuilder includes,
			final ServiceModule a, final ServiceModule b) {
		includes.append("#include <module.");
		includes.append(a.name);
		includes.append(">\n");
		includes.append("#include <module.");
		includes.append(b.name);
		includes.append(">\n");
		if (hasPointcuts(a)) {
			includes.append("#include <module.");
			includes.append(a.name);
			includes.append(".ah>\n");
		}
		if (hasPointcuts(b)) {
			includes.append("#include <module.");
			includes.append(b.name);
			includes.append(".ah>\n");
		}
	}
	
	private void generateAspect(final ServiceModule module) {
		StringBuilder content = new StringBuilder();
		StringBuilder includes = new StringBuilder();
		
		generateAspectIncludes(module, includes);
		
		content.append("\t/*** Pointcuts ***/\n");
		for (String pointcut : module.pointcuts) {
			content.append(pointcut);
			content.append("\n\n");
		}
		
		content.append("\t/*** Pointcuts only available at startup time ***/\n");
		for (String pointcut : module.startupPointcuts) {
			content.append(pointcut);
			content.append("\n\n");
		}
		
		content.append("\t/*** Pointcuts only available at shutdown time ***/\n");
		for (String pointcut : module.shutdownPointcuts) {
			content.append(pointcut);
			content.append("\n\n");
		}
				
		File outputFile = new File(outputBase, "module." + module.name + ".ah");
		String guard = getGuardFromName(module.name.getModule(), "ah");
		writeAspectFile(outputFile, guard, module.file.getName(), includes.toString(), module.name.getModule(), content.toString());
	}
	
	private void generateAspectIncludes(final ServiceModule module, final StringBuilder includes) {
		if (module.types.size() > 0) {
			includes.append("\n#include <module." + module.name + ".def.h>\n\n");
		}
		
		// Self include (of the own header file)
		includes.append("#include <module.");
		includes.append(module.name);
		includes.append(">\n");
		
		generateIncludes(module, includes);
		renderRequires(module, includes);
	}

	private void renderRequires(final ArchitectureElement element,
			final StringBuilder includes) {
		for (ServiceModule req : element.requirements) {
			if (hasPointcuts(req)) {
				includes.append("#include <module.");
				includes.append(req.name);
				includes.append(".ah>\n");
			}
		}
	}

	private void generateIncludes(final ServiceModule module, final StringBuilder includes) {
		if (module.types.size() > 0) {
			includes.append("\n#include <module." + module.name
					+ ".fwd.h>\n\n");
		}
		
		for (ServiceModule req : module.requirements) {
			includes.append("#include <module.");
			includes.append(req.name);
			includes.append(">\n");
		}
	}
	
	

	private void generateHeader(final ServiceModule module) {
		File outputFile = new File(outputBase, "module." + module.name);
		String guard = getGuardFromName(module.name.getModule(), "h");
		StringBuilder content = new StringBuilder();
				
		generateIncludes(module, content);
		
		for (String decl : module.declarations) {
			content.append(decl);
			content.append("\n");
		}
		
		if (module.types.size() > 0) {
			content.append("\n#include <module.");
			content.append(module.name);
			content.append(".def.h>\n\n");
		}
		
		writeHeaderFile(outputFile, guard, module.file.getName(),
				content.toString());
	}

	private String getGuardFromName(final String name, final String type) {
		return name.toUpperCase() + "_" + type.toUpperCase() + "_";
	}

	public final void initializeOdl() {
		odl.setSourceFolder(sourceDir);
		
		// extract implemented methods
		for (ServiceModule vmadlFile : serviceModules) {
			File file = new File(vmadlFile.file.getAbsolutePath().replace(".vmadl", ""));
			if (file.exists() && file.isDirectory()) {
				odl.analyze(file);
			}
		}

		odl.prepareCompilation();
		//odl.generateHeader(header);
		//odl.generateImplementation(implementation);
	}

	/**
	 * Prepare the module definitions for generation of files.
	 * 
	 * @throws CompilerException for any unexpected or wrong definition
	 */
	public final void compile() throws CompilerException {
		// parse type sections
	    // Set file and compile types section with ODL
	    for (ServiceModule module : serviceModules) {
	    	if (module.typesSection != null && module.typesSection != "") {
	    		parseTypesSection(module);
	    	}
	    }
		
		// compile replaces sections
		for (ServiceModule module : serviceModules) {
			for (Entry<QualifiedIdentifier, ServiceInterface> e
					: module.replacements.entrySet()) {
				ServiceModule m = reg.get(e.getKey().getModule());
				if (m.initialized) {
					m.parts.put(e.getKey(), e.getValue());
				}
			}
		}
		
		// flatten named parts
		for (ServiceModule module : serviceModules) {
			for (Entry<QualifiedIdentifier, ServiceInterface> e
					: module.parts.entrySet()) {
				ServiceInterface s = e.getValue();
				module.declarations.addAll(s.declarations);
				module.pointcuts.addAll(s.pointcuts);
				module.shutdownPointcuts.addAll(s.shutdownPointcuts);
				module.startupPointcuts.addAll(s.startupPointcuts);
				module.requirements.addAll(s.requirements);
				module.types.addAll(s.types);
			}
		}
		
		// revalidate consistency
		checkConsistency();
	}

	public final void loadFiles() {
		FilenameFilter filter = new FilenameFilter() { 
			public boolean accept(final File directory, final String filename) {
				return filename.endsWith(".vmadl"); }
		};
		File[] vmadlFiles = sourceDir.listFiles(filter);
		
		for (File file : vmadlFiles) {
			loadFromFile(file);
		}
		
	}

	public final void setUsedModules(final String[] modules) {
		serviceModules = reg.get(modules);
	}

	public final void setSourceDir(final String sourceDir) {
		this.sourceDir = new File(sourceDir);
	}

}
