/*
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
 */
 

#include <vm/Universe.h>
#include <vm/Image.h>


#include "memory.h"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>


#include <misc/Hashmap.h>
#include <misc/debug.h>


#include <vmobjects/objectformats.h>
#include <vmobjects/VMObject.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMMethod.h>
#include <vmobjects/VMBlock.h>
#include <vmobjects/VMPrimitive.h>
#include <vmobjects/VMEvaluationPrimitive.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/VMString.h>
#include <vmobjects/VMDouble.h>
#include <vmobjects/VMBigInteger.h>


#include <interpreter/Interpreter.h>
#include <compiler/GenerationContexts.h>

typedef struct _free_list_entry free_list_entry, *pFree_list_entry;
/*
 * free_list_entries are used to manage the space freed after the sweep phase.
 * these entries contain their own size and a reference of the entry next to them
 */
struct _free_list_entry {
    free_list_entry* next;
    size_t size;
};


/**
 *  \brief Header for ImageClass
 * 
 * Stores details on the heap, allowing for its contents to be restored
 * without loss of information.
 * 
 * Note, that all addresses are absolute (the raw pointers are stored). Thus
 * need to be updated when the heap is restored to memory and its new base address
 * is available.
 */ 
typedef struct _HeapHeader {
	INTEGER heap_size;			///< size of the heap in bytes
	void* old_base_address;	///< base address of the heap in memory at the moment of storage
	pVMFrame current_frame;		///< address to the currently active frame
	pFree_list_entry first_free_entry;	///< address of the first entry of the free list
	int size_of_free_heap;
} HeapHeader;


/*
 * heap used by gc_allocate
 */
void* object_space = NULL;


/*
 * the size of the heap used by gc_allocate (standard: 1 MB)
 */
size_t OBJECT_SPACE_SIZE = 1048576;


/*
 * will be adjusted when gc is initialized
 */
size_t BUFFERSIZE_FOR_UNINTERRUPTABLE = 50000;


/*
 * reference to the first entry in the free_list
 */
free_list_entry* first_free_entry = NULL;


/*
 * if this counter equals 0, only then it is safe to collect the
 * garbage. The counter is increased during initializations of
 * VMObjects and the generation of classes
 */
size_t uninterruptable_counter = 0;


size_t size_of_free_heap = 0;


//
// values for GC statistics
//


uint32_t num_collections; // the number of collections performed
uint32_t num_live;        // number of live objects (per collection)
uint32_t spc_live;        // space consumed by live objects (per collection)
uint32_t num_freed;       // number of freed objects (per collection)
uint32_t spc_freed;       // freed space (per collection)
uint32_t num_alloc;       // number of allocated objects (since last collection)
uint32_t spc_alloc;       // allocated space (since last collection)


//
// forward declarations
//

void gc_merge_free_spaces();
void gc_initialize();


void init_stat(void);
void init_collect_stat(void);
void collect_stat(void);
void reset_alloc_stat(void);


//
// implementations
//

// Tests if the object pointed to is located in the heap.
bool gc_is_managed_object(void* obj) {
	return obj >= object_space
		&& obj < (void*)((intptr_t)object_space + OBJECT_SPACE_SIZE);
}


void gc_set_heap_size(uint32_t heap_size) {
    // this can only be done before the heap is initialised - afterwards, it
    // yields an error message and terminates the VM
    if(object_space != NULL)
        Universe_error_exit("attempt to change heap size after initialisation");
    OBJECT_SPACE_SIZE = 1024 * 1024 * heap_size;
}

void gc_mark_reachable_stack_objects();

void gc_mark_reachable_objects() {
    // get globals
    pHashmap globals = (pHashmap) Universe_get_globals_dictionary();
    pHashmapElem elem = NULL;
    
    // iterate over the globals to mark all of them
	for(int i=0; i < globals->size; i++) {
        elem = (pHashmapElem) globals->elems[i];
        if (elem != NULL) {
            gc_mark_object(elem->key);
            gc_mark_object(elem->value);
        }
    }
	gc_mark_reachable_stack_objects();
}

void gc_mark_reachable_stack_objects() {
    // Get the current frame and mark it.
	// Since marking is done recursively, this automatically
	// marks the whole stack
    pVMFrame current_frame;
    Interpreter_get_frame(&current_frame);
    if (current_frame != NULL) {
        gc_mark_object(current_frame);
    }
}


/**
 *  check whether the object is inside the managed heap
 *  if it isn't, there is no VMObject to mark, otherwise
 *  check whether the object is already marked.
 *  if the self is an unmarked object inside the heap, then
 *  it is told to 'mark_references', recursively using this
 *  function for all its references.
 */
void gc_mark_object(void* _self) {
    pVMObject self = (pVMObject) _self;
    if (gc_is_managed_object(_self))
	{
        if (self->gc_field != 1) {
            // mark self before recursively marking contained references
            self->gc_field = 1;
            num_live++;
            spc_live += self->object_size;
            
            for (size_t i = 0; i < SEND(self, get_number_of_fields); i++) {
                gc_mark_object(SEND(self, get_field, i));
            }
        }
    }
}


/**
 * For debugging - the layout of the heap is shown.
 * This function uses 
 * - '[size]' for free space,
 * - '-xx-' for marked objects, and
 * - '-size-' for unmarked objects.
 * The output is also aligned to improve readability.
 */
void gc_show_memory() {
    pVMObject pointer = (pVMObject)object_space;
    free_list_entry* current_entry = first_free_entry;
    int object_size = 0;
    int object_aligner = 0;
    int line_count = 2;
    
    fprintf(stderr,"\n########\n# SHOW #\n########\n1 ");
    
    do {
        while (((void*)pointer > (void*)current_entry->next) && (current_entry->next != NULL)) {
            current_entry = current_entry->next;
        }
        
        if ((   (void*)current_entry->next == (void*)pointer) 
		    || ((void*)current_entry == (void*)pointer)) 
		{
            if ((void*)current_entry == (void*)pointer) {
                object_size = current_entry->size;
            } else {
                free_list_entry* next = current_entry->next;
                object_size = next->size;
            }
            fprintf(stderr,"[%d]",object_size);
        } else {
            pVMObject object = (pVMObject) pointer;
            object_size = SEND(object, object_size);

            if (object_size == 0) {
            	break;
            }
            
            // is this object marked or not?
            if (object->gc_field == 1) {
                fprintf(stderr,"-xx-");
            } else {
                fprintf(stderr,"-%d-", object_size);
            }
        }
		// aligns the output by inserting a line break after 36 objects
        object_aligner++;
        if (object_aligner == 36) {
            fprintf(stderr,"\n%d ", line_count++);
            object_aligner = 0;
        }
        pointer = (pVMObject)((intptr_t)pointer + object_size);
    } while ((void*)pointer < (void*)((intptr_t)object_space + OBJECT_SPACE_SIZE));    
}


void gc_collect() {
    num_collections++;
    init_collect_stat();
    
    if(gc_verbosity > 2) {
        fprintf(stderr, "-- pre-collection heap dump --\n");
        gc_show_memory();
    }
    
    gc_mark_reachable_objects();
    //gc_show_memory();
    pVMObject pointer = (pVMObject)object_space;
    free_list_entry* current_entry = first_free_entry;
    int object_size = 0;

    do {
        // we need to find the last free entry before the pointer
        // whose 'next' lies behind or is the pointer
        while (((void*)pointer > (void*)current_entry->next) && (current_entry->next != NULL)) {
            current_entry = current_entry->next;
        }
        
        if (((void*)current_entry->next == (void*)pointer) || ((void*)current_entry == (void*)pointer)) {
            // in case the pointer is the part of the free_list:
            if ((void*)current_entry == (void*)pointer) {
                object_size = current_entry->size;
            } else {
                free_list_entry* next = current_entry->next;
                object_size = next->size;
            }
            //fprintf(stderr,"[%d]",object_size);
            // nothing else to be done here
        } else {
            // in this case the pointer is a VMObject
            pVMObject object = pointer;
            object_size = SEND(object, object_size);

            if (object_size == 0) {
            	printf("Strange: Object_size == 0, try to escape here Pointer: %p\n", pointer);
            	break;
            }
            
            // is this object marked or not?
            if (object->gc_field == 1) {
                // remove the marking
				object->gc_field = 0;
            } else {
                num_freed++;
                spc_freed += object_size;
                
                // add new entry containing this object to the free_list
                SEND(object, free);
                memset(object, 0, object_size);
                free_list_entry* new_entry = (free_list_entry*) pointer;
                new_entry->size = object_size;

				// if the new entry lies before the first entry,
				// adjust the pointer to the first one
                if (new_entry < first_free_entry) {
                    new_entry->next = first_free_entry;
                    first_free_entry = new_entry;
                    current_entry = new_entry;
                } else {
                    // insert the newly created entry right after the current entry
                    new_entry->next = current_entry->next;
                    current_entry->next = new_entry;
                }
            }
        }
		// set the pointer to the next object in the heap
        pointer = (pVMObject)((intptr_t)pointer + object_size);

    } while ((void*)pointer < (void*)((intptr_t)object_space + OBJECT_SPACE_SIZE));

    // combine free_entries, which are next to each other
    gc_merge_free_spaces();
    
    if(gc_verbosity > 1)
        collect_stat();
    if(gc_verbosity > 2) {
        fprintf(stderr, "-- post-collection heap dump --\n");
        gc_show_memory();
    }
    
    reset_alloc_stat();
}


void* reserve_and_get_entry(pFree_list_entry entry, pFree_list_entry before_entry) {
	if (entry == first_free_entry) {
		// first one fitted - adjust the 'first-entry' pointer
		first_free_entry = entry->next;
	} else {
		// simply remove the reference to the found entry
		before_entry->next = entry->next;
	} // entry fitted
	return entry;
}


void split_and_reserve_entry(size_t size, pFree_list_entry entry, pFree_list_entry before_entry) {
	// save data from found entry
	int old_entry_size = entry->size;
	pFree_list_entry old_next = entry->next;

	// create new entry and assign data
	pFree_list_entry replace_entry =  (pFree_list_entry) ((intptr_t)entry + size);

	replace_entry->size = old_entry_size - size;
	if (replace_entry->size == 0) {
		printf("Strange: split seems to be wrong.\n");
	}

	replace_entry->next = old_next;
	if (entry == first_free_entry) {
		first_free_entry = replace_entry;
	} else {
		before_entry->next = replace_entry;
	}
}

void* memory_allocate(size_t size) {
    if(size == 0) return NULL;
	
	if(size < sizeof(struct _free_list_entry)) {
		return internal_allocate(size);
	}
    
    // first allocate-request => initialize heap
    if (object_space == NULL) {
        gc_initialize();
    }
    
    // start garbage collection if the free heap has less
	// than BUFFERSIZE_FOR_UNINTERRUPTABLE Bytes and this
	// allocation is interruptable
    if ((size_of_free_heap <= BUFFERSIZE_FOR_UNINTERRUPTABLE) && 
    		(uninterruptable_counter <= 0)) {
        gc_collect();
    }
    void* result = NULL;
    
    // initialize variables to search through the free_list
    free_list_entry* entry = first_free_entry;
    free_list_entry* before_entry = NULL;

	// don't look for the perfect match, but for the first-fit
    while (! ((entry->size == size) 
               || (entry->next == NULL) 
               || (entry->size >= (size + sizeof(struct _free_list_entry))))) { 
        before_entry = entry;
        entry = entry->next;
    }
    
	// did we find a perfect fit?
	// if so, we simply remove this entry from the list
    if (entry->size == size) {
        result = reserve_and_get_entry(entry, before_entry);
    } else {
		// did we find an entry big enough for the request and a new
		// free_entry?
        if (entry->size >= (size + sizeof(struct _free_list_entry))) {
            split_and_reserve_entry(size, entry, before_entry);
        	result = entry;
        }  else { 
			// no space was left
			// running the GC here will most certainly result in data loss!
			fprintf(stderr,"Not enough heap! Data loss is possible\n");
			fprintf(stderr, "FREE-Size: %d, uninterruptable_counter: %d\n",
				size_of_free_heap, uninterruptable_counter);
            
			gc_collect();
            //fulfill initial request
            result = memory_allocate(size);
        }
    }
           
    if(!result) {
        debug_error("Failed to allocate %d bytes. Panic.\n", (int)size);
        Universe_exit(-1);
    }
    memset(result, 0, size);
	
	// update the available size
    size_of_free_heap -= size;
    return result;
}


void* memory_allocate_object(size_t size) {
    size_t aligned_size = size + PAD_BYTES(size);
    void* o = memory_allocate(aligned_size);
    if(o)
        ((pOOObject)o)->object_size = aligned_size;
    num_alloc++;
    spc_alloc += aligned_size;
    return o;
}


/**
 * this function must not do anything, since the heap management
 * is done inside gc_collect.
 * However, it is called upon by all VMObjects.
 */
void memory_free(void* ptr) {
    // do nothing when called for an object inside the object_space
    if (!gc_is_managed_object(ptr))
	{
		internal_free(ptr);
	}
}


// free entries which are next to each other are merged into one entry
void gc_merge_free_spaces() {
    free_list_entry* entry = first_free_entry;
    free_list_entry* entry_to_append = NULL;
    int new_size = 0;
    free_list_entry* new_next = NULL;
    
    size_of_free_heap = 0;

    while (entry->next != NULL) {
        if (((intptr_t)entry + (intptr_t)(entry->size)) == (intptr_t)(entry->next)) {
            entry_to_append = entry->next;
            new_size = entry->size + entry_to_append->size;
            new_next = entry_to_append->next;
            
            memset(entry_to_append, 0, entry_to_append->size);
            entry->next = new_next;
            entry->size = new_size;
        } else {
			size_of_free_heap += entry->size;	
			entry = entry->next;
        }
    }
	if (entry->next == NULL) {
		size_of_free_heap += entry->size;	
	} else {
		fprintf(stderr, "Missed last free_entry of GC\n");
		Universe_exit(-1);
	}
}


void* internal_allocate(size_t size) {
    if(size == 0)
        return NULL;
    void* result = malloc(size);
    if(!result) {
        debug_error("Failed to allocate %d bytes. Panic.\n", (int)size);
        Universe_exit(-1);
    }
    memset(result, 0, size);
    return result;
}


void internal_free(void* ptr) {
    free(ptr);
}


char* internal_allocate_string(const char* restrict str) {
    return strdup(str);
}


/**
 * Sets up the heap and the free_list managing the free entries
 * inside the heap.
 */
void gc_initialize() { 
	// Buffersize is adjusted to the size of the heap (10%)
    BUFFERSIZE_FOR_UNINTERRUPTABLE = (int) (OBJECT_SPACE_SIZE * 0.1);

	// allocation of the heap
    object_space = malloc(OBJECT_SPACE_SIZE);
    if (!object_space) {
        fprintf(stderr, "Failed to allocate the initial %d bytes for the GC. Panic.\n",
                (int) OBJECT_SPACE_SIZE);
        Universe_exit(-1);
    } 
    memset(object_space, 0, OBJECT_SPACE_SIZE);
    size_of_free_heap = OBJECT_SPACE_SIZE;
    
    // initialize free_list by creating the first
    // entry, which contains the whole object_space
    first_free_entry = (free_list_entry*) object_space;
    first_free_entry->size = OBJECT_SPACE_SIZE;
    first_free_entry->next = NULL;
    
    // initialise statistical counters
    init_stat();
}


/**
 * if this counter is gt zero, an object initialization is
 * in progress. This means, that most certainlt, there are
 * objects, which are not reachable by the rootset, therefore,
 * the garbage collection would result in data loss and must
 * not be started!
 */
void gc_start_uninterruptable_allocation() {
    uninterruptable_counter++;
}


/**
 * only if the counter reaches zero, it is safe to start the
 * garbage collection.
 */
void gc_end_uninterruptable_allocation() {
    uninterruptable_counter--;
}


//
// functions for GC statistics and debugging output
//


/*
 * initialise statistics
 */
void init_stat(void) {
    num_collections = 0;

    // these two need to be initially set here, as they have to be preserved
    // across collections and cannot be reset in init_collect_stat()
    num_alloc = 0;
    spc_alloc = 0;
}


/*
 * initialise per-collection statistics
 */
void init_collect_stat(void) {
    // num_alloc and spc_alloc are not initialised here - they are reset after
    // the collection in reset_alloc_stat()
    num_live = 0;
    spc_live = 0;
    num_freed = 0;
    spc_freed = 0;
}


/*
 * reset allocation statistics
 */
void reset_alloc_stat(void) {
    num_alloc = 0;
    spc_alloc = 0;
}


#define _KB(B) (B/1024)
#define _MB(B) ((double)B/(1024.0*1024.0))


/*
 * Output garbage collection statistics. This function is called at the end of
 * a VM run.
 */
void gc_stat(void) {
    fprintf(stderr, "-- GC statistics --\n");
    fprintf(stderr, "* heap size %d B (%d kB, %.2f MB)\n",
        OBJECT_SPACE_SIZE, _KB(OBJECT_SPACE_SIZE), _MB(OBJECT_SPACE_SIZE));
    fprintf(stderr, "* performed %d collections\n", num_collections);
}


/*
 * output per-collection statistics
 */
void collect_stat(void) {
    fprintf(stderr, "\n[GC %d, %d alloc (%d kB), %d live (%d kB), %d freed "\
        "(%d kB)]\n",
        num_collections, num_alloc, _KB(spc_alloc), num_live, _KB(spc_live),
        num_freed, _KB(spc_freed));
}

/**
 * \brief Relocates the heap to a new memory position.
 * 
 * Allocates new memory for the heap, copies the contents and updates the object pointers.
 * This method allows to test an important part of the image functionality: The updating
 * of object references.  
 */
void gc_relocate_heap(void) {
	
	/// Set up the new heap
	debug_log("\n--------------------\n");
	debug_log("Starting to relocate heap.\n");
	
	// Save a pointer to the old heap
	void* old_heap = object_space;
	
	// Claim new memory, initialisation is not necessary since we overwrite it all anyways
    object_space = malloc(OBJECT_SPACE_SIZE);
    if (!object_space) {
    	debug_log("Failed to allocate %d bytes for relocating the heap. No relocation done.\n",
                (int) OBJECT_SPACE_SIZE);
        return;
    } 
    
    /// Perform the transition to the new heap.
    debug_log("New heap allocated, copying contents.\n");
    debug_log("[debug] Old heap @%xu -> @%xu\n", old_heap, ((uintptr_t)old_heap + OBJECT_SPACE_SIZE));
    debug_log("[debug] New heap @%xu -> @%xu\n", object_space, ((uintptr_t)object_space + OBJECT_SPACE_SIZE));
    memcpy(object_space, old_heap, (uintptr_t)OBJECT_SPACE_SIZE);
    
    debug_log("Updating free list.\n");
    
    // Calculate the difference of the pointers, which is all that is need for 
    // updating the pointers.
    /// \note This may cause problems in Win32 if the difference is larger than the value range
    // for a int32.
    int difference = (INTEGER)object_space - (INTEGER)old_heap;
    void* new_pointer = (void*)((uintptr_t)old_heap + difference);
    if (object_space != new_pointer) {
    	// Adding the difference did not produce the expected result.
    	// This probably means that the difference between the heaps is larger than the value
    	// range of a signed int. 
    	
    	// Clean up new heap and restore the old one.
    	free(object_space);
    	object_space = old_heap;
    	
    	// continue (hopefully) normally
    	debug_log("Error relocating heap (int probably was not big enough).");
    	debug_log("Relocation aborted. Continuing.");
    	return;
    }
    
    // update the free_list pointers, starting with the first entry
    first_free_entry = (pFree_list_entry)((INTEGER)(first_free_entry) + difference);
    pFree_list_entry free_entry = first_free_entry;
    while (free_entry->next) {
    	free_entry->next = (pFree_list_entry)((INTEGER)(free_entry->next) + difference);
    	free_entry = free_entry->next;
    }
    
	// Get a pointer to the current frame
    debug_log("Updating Universe-related references.\n");
    
    // Update the globals
    Universe_update_references(difference);

    debug_log("Updating frame references.\n");
    
    // Update all pointers on the interpreter stack, thus starting with the topmost frame.
    // This task is object-specific, so the responsibility is delegated
    // to the objects themselves.
    pVMFrame c;
    Interpreter_get_frame(&c);
    pVMFrame current_frame = (pVMFrame)((INTEGER)c + difference);
    Interpreter_set_frame(current_frame);
    SEND(current_frame, update_references, difference);
    
    /// Get rid of the old heap
    debug_log("Remove the old heap.\n");
    
    // First zero the memory out to trigger NULL-pointer seg-faults,
    // if the old heap is still accessed and then free it.
    memset(old_heap, 0, OBJECT_SPACE_SIZE);
    free(old_heap);
	
    debug_log("Done.\n");
    debug_log("--------------------\n");
}


// Serializes the state of the GC, especially the heap,
// to the Image component.
bool gc_serialize_heap() {

	// store heap header information
	HeapHeader hh;
	hh.heap_size = OBJECT_SPACE_SIZE;
	Interpreter_get_frame(&hh.current_frame);
	hh.first_free_entry = first_free_entry;
	hh.old_base_address = object_space;
	hh.size_of_free_heap = size_of_free_heap;
	
	if (!image_store_byte_sequence(sizeof(hh), &hh))
		return false;
	
	// store heap
	if (!image_store_byte_sequence(OBJECT_SPACE_SIZE, object_space))
		return false;
	
	return true;
}

// Restores the GC state from the Image component,
// especially restoring the heap.
bool gc_deserialize_heap(INTEGER* pdiff) {	
	debug_trace("old heap area: %xu<->%xu\n", object_space, (void*)((INTEGER)object_space + (INTEGER)OBJECT_SPACE_SIZE));

	INTEGER header_size, heap_size;

	// load header
	HeapHeader* phh = (HeapHeader*)image_load_byte_sequence(&header_size, false);
	if (!phh || (sizeof(HeapHeader) != header_size))
		return false;
	
	// copying struct contents, allowing buffer to be freed on next call to image_load_byte_sequence
	HeapHeader hh = *phh;

	// free old heap (if necessary) and load new one
	if (object_space) {
	    memset(object_space, 0, OBJECT_SPACE_SIZE);
	    free(object_space);
	}
	object_space = image_load_byte_sequence(&heap_size, true);

	if (!object_space)
		return false;
	
	debug_trace("new heap area: %xu<->%xu\n", object_space, (void*)((INTEGER)object_space + (INTEGER)OBJECT_SPACE_SIZE));

	if (hh.heap_size != heap_size)
	{
		free(object_space);
		return false;
	}

	// utilizing info from header to update GC-local values
	OBJECT_SPACE_SIZE = hh.heap_size;
	size_of_free_heap = hh.size_of_free_heap;

	// save return value and get a pure INTEGER to avoid non-dereferenced-pointer mistakes
    *pdiff = (INTEGER)object_space - (INTEGER)(hh.old_base_address);
    INTEGER diff = *pdiff;
    debug_trace("diff in heap addresses: %x\n", diff);
    
	// update the free_list pointers, starting with the first entry
    first_free_entry           = (pFree_list_entry)((INTEGER)(hh.first_free_entry) + diff);
    pFree_list_entry free_entry = first_free_entry;
    while (free_entry->next) {
    	free_entry->next = (pFree_list_entry)((INTEGER)free_entry->next + diff);
    	free_entry = free_entry->next;
    }
    
    // Update all pointers on the interpreter stack, thus starting with the topmost frame.
    // This task is object-specific, so the responsibility is delegated
    // to the objects themselves.
	pVMFrame current_frame     = (pVMFrame)((INTEGER)(hh.current_frame) + diff);
    update_object_references((pVMObject)current_frame, diff);
	
	return true;
}


// serializes object-specific data that is not in the heap
bool gc_serialize_object_data(void) {
	pVMFrame current_frame;
        Interpreter_get_frame(&current_frame);
	return serialize_object_data((pVMObject)current_frame);
}

// loads object-specific data that is not in the heap 
bool gc_deserialize_object_data(void) {
	pVMFrame current_frame;
        Interpreter_get_frame(&current_frame);
	return deserialize_object_data((pVMObject)current_frame);
}
