#ifndef MEMORY_H_
#define MEMORY_H_

/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
 */
 
#include <vmobjects/VMObject.h>
#include <compiler/GenerationContexts.h>

#include <stdlib.h>
#include <misc/defs.h>


/*
 * macro for padding - only word-aligned memory must be allocated
 */
#define PAD_BYTES(N) ((sizeof(void*) - ((N) % sizeof(void*))) % sizeof(void*))


/*
 * The heap size can be set on VM startup. The argument passed to this function
 * represents the dedicated heap size in MB.
 */
void gc_set_heap_size(uint32_t heap_size);


void gc_mark_object(void* _self);


void gc_collect();
void gc_start_uninterruptable_allocation();
void gc_end_uninterruptable_allocation();


void*  memory_allocate(size_t size);
void*  memory_allocate_object(size_t size);
char*  memory_allocate_string(const char* restrict str);
void   memory_free(void* ptr);

void gc_stat(void);

/**
 * \brief Relocates the heap at run-time to a new memory location.
 * 
 * Copies the heap and makes sure, that the new copy is used from now on.
 * Then the old heap is freed.
 * 
 * This method has been created solely for preparing the image
 * functionality, but can still be accessed through the "system rf"
 * command. (The rf message is obviously implemented as a primitive,
 * pointing here.)
 */
void gc_relocate_heap(void);

/**
 * \brief Tests if the object is located in the heap.
 * 
 * Tests whether the pointed to object provided as parameter is
 * located on the heap. This is accomplished by testing whether
 * the address is within the range of the heap memory area.
 * 
 * \param obj The object to be tested.
 * 
 * \return Whether or not the object lies in the object heap.
 */
bool   gc_is_managed_object(void* obj);

void*  internal_allocate(size_t size);
char*  internal_allocate_string(const char* restrict str);
void   internal_free(void* ptr);

/**
 * \brief GC-managed data is serialized into image file.
 * 
 * The garbage collector serializes all its data, including
 * the heap, into the image file being currently written.
 * 
 * Serializing occurs via the image call-back functions, so
 * the image must take care of deserializing taking place in
 * the same fashion that serializing did.
 * 
 * \return Whether the data could be serialized successfully.
 */
bool gc_serialize_heap(void);

/**
 * \brief Retrieves GC-managed data from the image file.
 * 
 * The garbage collector retrieves its data from the image
 * file, using again the callback functions.
 * 
 * \return Whether deserializing worked.
 */
bool gc_deserialize_heap(INTEGER* diff);

/**
 * \brief Make the objects save their own data in the image file.
 * 
 * Initiates the process during which the GC managed objects
 * store their own heap-external data in the image file.
 */
bool gc_serialize_object_data(void);

/**
 * \brief Make the objects retrieve their specific data from the image file.
 * 
 * Initiates the process during which the GC managed objects
 * retrieve their data that is not stored in the heap.
 * 
 * \note Only after this function returns are symbols and strings available.
 * This is usually important for debugging where one for example might want
 * to see the human-readable class name.
 */
bool gc_deserialize_object_data(void);

#endif MEMORY_H_
