/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

/**
 * \file Image.c
 * \brief Implementation of image functionality.
 * 
 * This file contains the implementation of the image functionality as
 * far as the component-independent data is concerned.
 * By calling the components to save their data, this Image component
 * stays clear of much VM implementation details.
 * 
 * At the same time,
 * the components were not supposed to bother with implementation
 * of the image functionality. For this purpose call-back functions
 * are implemented to allow for arbitrary byte sequences to be stored,
 * and all that matters is the order in which to ask the components
 * to (de)serialize their data. (Note that someone must always know
 * what data is needed where and how to interpret loaded data. So this
 * someone may as well be the corresponding component, as long as it is
 * not bothered with IO-details and other non data structure related
 * issues.
 */

#include "Image.h"

//required for compression
#include <zlib.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>

// VM parts used for hibernation and wakeup procedures
#include <vm/Universe.h>
#include <misc/debug.h>

#include <interpreter/Interpreter.h>


#include <vmobjects/VMArray.h>
#include <vmobjects/VMBigInteger.h>
#include <vmobjects/VMBlock.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/VMDouble.h>
#include <vmobjects/VMEvaluationPrimitive.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMImage.h>
#include <vmobjects/VMInteger.h>
#include <vmobjects/VMInvokable.h>
#include <vmobjects/VMMethod.h>
#include <vmobjects/VMObject.h>
#include <vmobjects/VMPrimitive.h>
#include <vmobjects/VMString.h>
#include <vmobjects/VMSymbol.h>

// Garbage collector
#include <memory/memory.h>


/// \brief The file ending of all image files, automatically appended
static const char* fname_ending = ".cim";
static const char* temp_img_ending = ".tmp";
static const int MAX_FNAME_SIZE = 150; ///< max size of file name buffer

/// \brief Image being currently (re)stored, NULL if neither is in progress.
static FILE* image_file = NULL;

/// \brief Image header of the image which is currently loading/saving
static ImageHeader image_header;

/**
 *  \brief Whether to free or retain dynamic buffer.
 * 
 * Answers the question regarding memory ownership.
 * (Me longs for smart pointers.)
 */ 
static bool retain_dynamic_buffer;
static void* dynamic_buffer; ///< The dynamic buffer from the callbacks.


//////////////////////////////////////////////////////////
/// structs and types

typedef struct _VMT_Addresses {
	VTABLE(VMArray)*               vmt_array;
	VTABLE(VMBigInteger)*          vmt_big_integer;
	VTABLE(VMBlock)*               vmt_block;
	VTABLE(VMClass)*               vmt_class;
	VTABLE(VMDouble)*              vmt_double;
	VTABLE(VMEvaluationPrimitive)* vmt_evaluation_primitive;
	VTABLE(VMFrame)*               vmt_frame;
	VTABLE(VMImage)*               vmt_image;
	VTABLE(VMInteger)*             vmt_integer;
	VTABLE(VMInvokable)*           vmt_invokable;
	VTABLE(VMMethod)*              vmt_method;
	VTABLE(VMObject)*              vmt_object;
	VTABLE(VMPrimitive)*           vmt_primitive;
	VTABLE(VMString)*              vmt_string;
	VTABLE(VMSymbol)*              vmt_symbol;
} VMT_Addresses;

static VMT_Addresses   current_vmt_pointers; ///< currently valid VMT pointers
/// \brief whether current_vmt_pointers is initialized with valid values
static bool            current_vmt_pointers_initialized = false;
/// \brief VMT pointers retrieved from the loaded image.
/// \note This obviously is only interesting, as long as the loading is in
/// progress and becomes superfluous the moment all VMTs have been updated.
static VMT_Addresses   old_vmt_pointers;

//////////////////////////////////////////////////////////
/// Forward declarations of helper functions

/**
 * \brief Aborts file reading during loading.
 * 
 * Performs all necessary cleanup steps to guarantee a clean
 * abort of image loading. This concerns mainly the internal
 * state of this Image component. 
 */   
void abort_image_loading(void);
/**
 * \brief Aborts file writing during storing.
 * 
 * Performs all necessary cleanup steps to guarantee a clean
 * abort of image storing. Besides closing the file pointer,
 * the file is also deleted from the disc, since its empty
 * at best but most likely corrupted. 
 * 
 * \param file_name The name of the image file (to be deleted).
 */   
void abort_image_storing(const char* const file_name);

/// \brief Constructs a canonical filename from a given image name, currently "[image_name].cim"
bool construct_image_filename(const pString image_name, char* const image_fname);

/// \brief Load the VMT pointers from an image and store it in the file-local struct. 
bool deserialize_vmt_pointers(void);

/**
 * \brief Tests if an image is loadable.
 * 
 * \param image_header The ImageHeader from the image file.
 * \return Whether the image can be loaded according to its current header.
 */  
bool is_loadable(const ImageHeader image_header);

/**
 * \brief Frees memory for a dynamic buffer.
 * 
 * Memory used by a callback is freed, if the caller has not specified
 * for it to be retained.
 */
void free_dynamic_buffer(void);

/** \brief Actually loads the image with the given name
 * 
 * \param image_fname Path and filename of the image to load.
 * \return Whether image loading worked. If not the system cannot be
 *       expected to be able to continue (and it probably won't).
 */
bool load_image(pString image_fname, const int32_t encrypt_key);

/// \brief Serializes the currently VMT Pointers using the callback functions.
bool serialize_vmt_pointers(void);

/// \brief Initializes the struct containing the currently valid VMT pointers.
void set_current_vmt_pointers(void);

/**
 * \brief Sets the image header entries to their current constants.
 * \param h The header to update.
 */
void set_image_header_constants(ImageHeader* const h);

/**
 * \brief Reads the image header and makes it globally
 * accessible
 */
bool read_image_header(const char* image_fname);

/**
 * \brief Compress function: Compresses the given data stream.
 * 
 * \todo document me
 * 
 * \param size Number of bytes to be stored from byte_sequence.
 * \param byte_sequence The data to store. 
 */
bool compress_image_file(const char* image_fname);
bool compress_data(const INTEGER size, const void* byte_sequence, INTEGER* new_size, void** new_byte_sequence);

/**
 * \brief Decompress function: Decompresses the given data stream.
 * 
 * \todo document me
 * 
 * \param size Number of bytes to be stored from byte_sequence.
 */
bool decompress_image_file(const char* image_fname);
bool decompress_data(const INTEGER size, const void* byte_sequence, INTEGER* new_size, void** new_byte_sequence);

/**
 * \brief Encrypts the image file. Therefore the stored image is loaded,
 * the complete file (without the image header) will be encrypted. After
 * encryption the image header and an ImageEncryptionHeader is written
 * to the image. After the two headers the encrypted image follows.
 * \note The encryption method is stored in the ImageEncryptionHeader   
 */
bool encrypt_image_file(const char* image_fname, int32_t encryption_method, const void* key);

/** \brief Decrypts the image file. Therefore the the stored image is loaded,
 * the ImageEncryptionHeader is read and the image will be decrypted. After
 * decryption the image is stored unencrypted and without ImageEncryptionHeader.
 */
bool decrypt_image_file(const char* image_fname, const void* key);


//////////////////////////////////////////////////////////
/// Macrodefinitions
/**
 * Macro for better maintainability, this reads the first header, while
 * sciping the ImageHeader
 */
#define READ_HEADER(BUFFER, SIZE, IMAGE_FILE) \
	   (   0 == fseek(IMAGE_FILE, sizeof(ImageHeader), SEEK_SET)\
		&& 1 == fread(BUFFER, SIZE, 1, IMAGE_FILE))

/**
 * Image Format Info Bits
 * Declares the Format of the Image and the contained headers
 * can have values like: 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 
 * 1024, 2048, 4096, 8192, ..., 2^31
 */
#define FORMAT_COMPRESSION_BIT	1
#define FORMAT_ENCRYPTION_BIT 	2
/**
 * ImageFormatInfoMacros
 */
/// \brief Sets the format bit
#define SET_FORMAT_BIT(IMAGE_FORMAT_INFO, FORMAT_BIT)\
	IMAGE_FORMAT_INFO = IMAGE_FORMAT_INFO | FORMAT_BIT

/// \brief Checks if the format bit is set
#define CHECK_FORMAT_BIT(IMAGE_FORMAT_INFO, FORMAT_BIT)\
	(FORMAT_BIT & IMAGE_FORMAT_INFO)

/**
 * \brief Copies file src to file dest
 */
bool fcopy(const char *source, const char *dest);

//////////////////////////////////////////////////////////
/// Public functions

// The currently running image is stored in file. 
bool image_store(const pString image_name, const int32_t encrypt_alg, const int32_t encrypt_key) {
	
	/// If image storing is currently in progress, abort.
	if (image_file)
		return false;
	
	/// \bug please test me
	/// (we had a situation, where gc_collect() crashed after 3 invocations).
	// still, running the GC is necessary to remove objects (and object's non heap data!!)
	// so that restoration goes smoothly.
	gc_collect();
	
	// Construct filename
	char image_fname[ MAX_FNAME_SIZE ];
	if (!construct_image_filename(image_name, image_fname))
		return false;
	debug_log("Storing image in: '%s'\n", image_fname);
	
	// Open file
	image_file = fopen(image_fname, "wb");
	if (!image_file)
		return false;
	
	// store image header
	set_image_header_constants(&image_header);
	
	if (1 != fwrite(&image_header, sizeof(image_header), 1, image_file)) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// store info header, currently with dummy values (whatever is in memory)
	ImageInfoHeader info_header;
	if (1 != fwrite(&info_header, sizeof(info_header), 1, image_file)) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// store pointers to VMTs
	info_header.vmt_data_offset = ftell(image_file);
	if (!serialize_vmt_pointers()) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// store heap
	info_header.gc_data_offset = ftell(image_file);
	if (!gc_serialize_heap()) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// store Universe state
	info_header.universe_data_offset = ftell(image_file);
	if (!Universe_serialize()) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// store Interpreter state
	info_header.interpreter_data_offset = ftell(image_file);
	if (!Interpreter_serialize()) {
		abort_image_storing(image_fname);
		return false;
	}
	
	/////////////// Serialize object-specific data
	/// \note this may be removed later on, as object-specific data is all in the heap
	
	if (!gc_serialize_object_data()) {
		abort_image_storing(image_fname);
		return false;
	}
	
	if (!Universe_serialize_object_data()) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// Write filled info header.
	info_header.timestamp_of_storage = time(NULL);
	if (   0 != fseek (image_file, sizeof(image_header), SEEK_SET      )
		|| 1 != fwrite(&info_header, sizeof(info_header), 1, image_file) ) {
		abort_image_storing(image_fname);
		return false;
	}
	
	// close file
	if (0 != fclose(image_file)) {
		abort_image_storing(image_fname);
		return false;
	}

//	// Compress image
//	if (!compress_image_file(image_fname)) {
//		abort_image_storing(image_fname);
//		return false;
//	}
	
	// Encrypt image, if wanted
	if (IMAGE_CRYPT_NONE != encrypt_alg)
		if (!encrypt_image_file(image_fname, encrypt_alg, &encrypt_key)) {
			abort_image_storing(image_fname);
			return false;
		}

	image_file = NULL;
	return true;
}

// The given file is inspected for possible validity.
pVMImage image_inspect(const pString image_name) {
	
	// abort, if image operation is currently in progress
	if (image_file)
		return NULL;
	
	// open file
	char image_fname[ MAX_FNAME_SIZE ];
	if (!construct_image_filename(image_name, image_fname))
		return NULL;
	
	image_file = fopen(image_fname, "rb");
	if (!image_file) {
		fprintf(stderr, "Cannot load %s\n", image_fname);
		return NULL;
	}
	
	// variables to read file contents into
	ImageHeader image_header;
	ImageInfoHeader info_header;
	
	// read the two headers
	// note: reading of the info header should work independent of semantic or syntactic
	// incompatibilities.
	if (     1 != fread(&image_header, sizeof(image_header), 1, image_file)
	      || 1 != fread(&info_header, sizeof(info_header), 1, image_file))
		return NULL;
	
	if (0 != fclose(image_file)) {
		// do nothing
	}
	
	// If the image is compressed, encrypted or otherwise not readable,
	// the data in the info header is nonsense and is deleted (with zeros).
	if (0 != image_header.format_info)
		memset(&info_header, 0, sizeof(info_header));
		
	
	// determine file size
	struct stat buf;
	stat(image_fname, &buf);
	int file_size = buf.st_size;
	
	// create VMImage instance
	gc_start_uninterruptable_allocation();
	
	pString ps = String_new(image_fname);
	pVMString vm_fname = VMString_new_with(ps);
	pVMInteger vm_fsize = VMInteger_new_with(file_size);
	pVMImage img = Universe_new_image(vm_fsize, vm_fname, image_header, info_header);
	
	gc_end_uninterruptable_allocation();
	
	return img;
}

// Tests whether the given image is compressed
bool image_is_compressed(pVMImage image) {
	
	ImageHeader image_header = SEND(image, get_image_header);

	return CHECK_FORMAT_BIT(image_header.format_info, FORMAT_COMPRESSION_BIT);
}

// Tests whether the given image is encrypted
bool image_is_encrypted(pVMImage image) {
	
	ImageHeader image_header = SEND(image, get_image_header);

	return CHECK_FORMAT_BIT(image_header.format_info, FORMAT_ENCRYPTION_BIT);
}

// Tests whether the given image is loadable (judging from its headers).
bool image_is_loadable(pVMImage image) {
	
	ImageHeader image_header = SEND(image, get_image_header);

	return is_loadable(image_header);
}

// Restores the image with the given name.
bool image_load_with_name(pString image_name, const int32_t encrypt_key) {
	char fname[ MAX_FNAME_SIZE ];
	
	if (!construct_image_filename(image_name, fname))
		return false;
	
	return load_image(String_new(&(fname[0])),  encrypt_key);
}

// Loads the given image and resumes its execution.
bool image_load(pVMImage image, const int32_t encrypt_key) {
	
	// Note: All that is on the heap looses its validity after image loading has begun.
	// This includes arguments of this very function, so they must be copied
	// to non-VM-heap memory, if they are needed afterwards.
	
	if (!image_file)
		return false;
	
	if (!is_loadable(SEND(image, get_image_header)))
		return false;
	
	// open file
	pString image_fname = SEND(SEND(image, get_file_name), get_embedded_string);
	
	return load_image(image_fname, encrypt_key);
}

// Callback function for storing caller-specific data. 
void* image_load_byte_sequence(INTEGER* size, bool retain) {
	
	/// \todo implement boundary check by taking advantage of the offsets in the info header
	
	if (!image_file || feof(image_file))
		Universe_error_exit("Image callback function was called, but image loading was not in progress or EOF is reached.");
	
	// free buffers from previous call
	free_dynamic_buffer();
	
	// load how large this chunk is
	INTEGER bytes_to_read;
	if (1 != fread(&bytes_to_read, sizeof(INTEGER), 1, image_file))
		return NULL;
	
	// return size to caller, if he is interested in
	if (size)
		*size = bytes_to_read;
	
	// prepare buffer
	retain_dynamic_buffer = retain;
	dynamic_buffer = malloc(bytes_to_read);
	if (!dynamic_buffer)
		return NULL;
	
	// read contents
	if (1 != fread(dynamic_buffer, bytes_to_read, 1, image_file))
		return NULL;
	
	return dynamic_buffer;
}

// Callback function for storing caller-specific data. 
bool image_store_byte_sequence(const INTEGER size, const void* const byte_sequence) {
	
	if (!image_file)
		Universe_error_exit("Image callback function was called, but image storing was not in progress.");
	
	// save how large this chunk is
	if (1 != fwrite(&size, sizeof(INTEGER), 1, image_file))
		return false;
	
	// write contents
	if (1 != fwrite(byte_sequence, size, 1, image_file))
		return false;
	
	return true;
}


// translates a vmt pointer from a loaded image to the currently valid pointer
void* image_translate_vmt_pointer(const void* const old) {
	
	set_current_vmt_pointers();
	
	VMT_Addresses oldp = old_vmt_pointers;
	VMT_Addresses newp = current_vmt_pointers;
	
	int items = sizeof(VMT_Addresses) / sizeof(void*);
	for (int curr = 0; curr < items; curr++)
		if (old == ((void**)&old_vmt_pointers)[ curr ]) 
			return ((void**)&current_vmt_pointers)[ curr ];

	debug_trace("Could not translate VMT pointer (%x)\n", old);
	Universe_error_exit("Could not translate VMT pointer, dying happily.");

	return NULL; // for completeness sake
}

//////////////////////////////////////////////////////////
/// Private functions


// Cleans up when image loading fails.
void abort_image_loading() {
	
	// free buffers
	free_dynamic_buffer();
	
	// close the file, set the pointer to NULL 
	old_vmt_pointers = current_vmt_pointers;
	fclose(image_file);
	image_file = NULL;
}

// Cleans up, when image storing fails in the process.
void abort_image_storing(const char* const file_name) {
	
	// close the file, set the pointer to NULL
	abort_image_loading();
	
	// additionally delete the file, if any
	remove(file_name);
}

// Creates a canonical file name for a given image name. 
bool construct_image_filename(const pString image_name, char* const image_fname) {
	int emb_str_len = SEND(image_name, length);
	int fname_ending_len = strlen(fname_ending);
	int temp_ending_len = strlen(temp_img_ending);

	// larger than buffer?
	if (MAX_FNAME_SIZE < emb_str_len + fname_ending_len + temp_ending_len + 1) // filename + ending + temp_img_ending + \0
		return false;
	const char * const raw_image_name = SEND(image_name, chars);
	
	// copy string + ending + terminating null
	int idx = 0;
	strncpy(&(image_fname[ idx ]), raw_image_name, emb_str_len); idx += emb_str_len;
	strncpy(&(image_fname[ idx ]), fname_ending, fname_ending_len); idx += fname_ending_len;
		
	image_fname[ idx ] = '\0';
	
	return true;
}

// Retrieves the (now probably invalid) VMT pointers from the opened image.
bool deserialize_vmt_pointers(void) {
	
	// load old vmt pointers and save them
	INTEGER bytes_read;
	VMT_Addresses* tmp_vmt_pointer = (VMT_Addresses*)image_load_byte_sequence(&bytes_read, false);
	if (!tmp_vmt_pointer || sizeof(old_vmt_pointers) != bytes_read)
		return false;
	
	old_vmt_pointers = *tmp_vmt_pointer;
	
	set_current_vmt_pointers();
	
	return true;
}


// Frees the dynamic buffer used by the callback function to
// retrieve data from an image.
void free_dynamic_buffer(void) {
	
	if (!dynamic_buffer || retain_dynamic_buffer)
		return;
	
	free(dynamic_buffer);
	dynamic_buffer = NULL;
}


// tests whether an image_header indicates a compatible
// (and subsequently a loadable) image
bool is_loadable(const ImageHeader image_header) {
	
	// test magic constant
	if (IMAGE_MAGIC_CONSTANT != image_header.magic)
		return false;
	
	// test version, minor version should not be of relevance
	if (IMAGE_VERSION != image_header.version)
		return false;
	
	// test if word-length-architecture fits (i.e. 32 vs 64 bit)
	if ((sizeof(void*) * 8) != image_header.word_length)
		return false;
	
	return true;
}

/// creates the name for the temporary image file
void create_temp_image_name(const char* org_fname, char* new_fname) {
	int fname_org_len = strlen(org_fname);
	int temp_img_ending_len = strlen(temp_img_ending);
	
	// larger than buffer?
	if (MAX_FNAME_SIZE < fname_org_len + temp_img_ending_len + 1) // filename + ending + \0
		return;
	
	// copy string + ending + terminating null
	int idx = 0;
	strncpy(&(new_fname[ idx ]), org_fname, fname_org_len); idx += fname_org_len;
	strncpy(&(new_fname[ idx ]), temp_img_ending, temp_img_ending_len); idx += temp_img_ending_len;
		
	new_fname[ idx ] = '\0';
}

// Actually loads the image with the given name.
bool load_image(pString image_fname, const int32_t encrypt_key) {
	const char* fname_org = SEND(image_fname, chars);

	// create new name for temporary image ...
	char fname[MAX_FNAME_SIZE];
	create_temp_image_name(fname_org, &(fname[0]));
	
	// and copy the original image file
	if (!fcopy(fname_org, fname))
		return false;
	
	read_image_header(fname);
	
	// test loadability
	if (!is_loadable(image_header))
		return false;
	
	// decrypt the image if it is encrypted
	if (CHECK_FORMAT_BIT(image_header.format_info, FORMAT_ENCRYPTION_BIT)) {
		if (!decrypt_image_file(fname, &encrypt_key)) {
			abort_image_loading();
			return false;
		}
	}
	
	// decompress the image if it is compressed
//	if (CHECK_FORMAT_BIT(image_header.format_info, FORMAT_COMPRESSION_BIT)) {
//	
//		// Decompress image
//		if (!decompress_image_file(fname)) {
//			abort_image_loading();
//			return false;
//		}
//
//	}
	
	image_file = fopen(fname, "rb");
	if (!image_file) {
		fprintf(stderr, "Cannot load %s\n", fname);
		return false;
	}
	
	// variables to read file contents into
	ImageInfoHeader info_header;
	
	// note: reading of the info header should work independent of semantic or syntactic
	// incompatibilities.
	if (!READ_HEADER(&info_header, sizeof(info_header), image_file))
		return false;
	
	// prepare dynamic memory
	dynamic_buffer = NULL;
	retain_dynamic_buffer = false;

	/////////////// Deserialize 
	
	// restore heap
	free_dynamic_buffer();
	set_current_vmt_pointers();
	
	if (!deserialize_vmt_pointers()) {
		abort_image_loading();
		return false;
	}
	
	// restore heap
	INTEGER diff;
	free_dynamic_buffer();
	if (!gc_deserialize_heap(&diff)) {
		abort_image_loading();
		return false;
	}
	
	// restore universe data
	free_dynamic_buffer();
	if (!Universe_deserialize(diff)) {
		abort_image_loading();
		return false;
	}
	
	// restore interpreter data
	free_dynamic_buffer();
	if (!Interpreter_deserialize(diff)) {
		abort_image_loading();
		return false;
	}
	
	/////////////// Deserialize object-specific data
	/// \note this may be removed later on, as object-specific data is all in the heap
	if (!gc_deserialize_object_data()) {
		abort_image_loading();
		return false;
	}

	if (!Universe_deserialize_object_data()) {
		abort_image_loading();
		return false;
	}

	
	/////////////// Finalizing Universe
	/// resets the Universe globals and restores the primitives
	if (!Universe_finalize(diff)) {
		abort_image_loading();
		return false;
	}
	
	// close file
	if (0 != fclose(image_file))
		// returning true, because aborting at this point doesn't make any sense
		debug_warn("Could not close image file, but continuing happily. Please do test the image.");
	
	// cleanup
	free_dynamic_buffer();
	image_file = NULL;
	old_vmt_pointers = current_vmt_pointers;
	
	return true;
}


// serialize currently valid VMT pointers
bool serialize_vmt_pointers(void) {
	set_current_vmt_pointers();
	
	if (!image_store_byte_sequence(sizeof(current_vmt_pointers), &current_vmt_pointers))
		return false;

	return true;
}


// set the currently valid VMT pointers in the file-local struct
void set_current_vmt_pointers() {
	
	if (current_vmt_pointers_initialized)
		return;
	
	current_vmt_pointers.vmt_array                = VMArray_vtable();
	current_vmt_pointers.vmt_big_integer          = VMBigInteger_vtable();
	current_vmt_pointers.vmt_block                = VMBlock_vtable();
	current_vmt_pointers.vmt_class                = VMClass_vtable();
	current_vmt_pointers.vmt_double               = VMDouble_vtable();
	current_vmt_pointers.vmt_evaluation_primitive = VMEvaluationPrimitive_vtable();
	current_vmt_pointers.vmt_frame                = VMFrame_vtable();
	current_vmt_pointers.vmt_image                = VMImage_vtable();
	current_vmt_pointers.vmt_integer              = VMInteger_vtable();
	current_vmt_pointers.vmt_invokable            = VMInvokable_vtable();
	current_vmt_pointers.vmt_method               = VMMethod_vtable();
	current_vmt_pointers.vmt_object               = VMObject_vtable();
	current_vmt_pointers.vmt_primitive            = VMPrimitive_vtable();
	current_vmt_pointers.vmt_string               = VMString_vtable();
	current_vmt_pointers.vmt_symbol               = VMSymbol_vtable();
	
	current_vmt_pointers_initialized = true;
	
	old_vmt_pointers = current_vmt_pointers;
	
}

// Fills the ImageHeader struct with currently applicable values.
void set_image_header_constants(ImageHeader* const h) {
	h->magic = IMAGE_MAGIC_CONSTANT;
	h->version = IMAGE_VERSION;
	h->word_length   = (sizeof(void*) * 8);
	
	// can be set via SET_FORMAT_BIT Macro
	h->format_info = 0;
}

// reads the image header
bool read_image_header(const char* image_fname) {
	// Open file for reading
	image_file = fopen(image_fname, "rb");
	if (!image_file)
		return false;
	
	ImageHeader ih;
	if (1 != fread(&image_header, sizeof(ImageHeader), 1, image_file))
		return false;
	
	return true;
}

/** 
 * \bug not completely implemented
 * decompresses the given byte_sequence and returns it by conpressed data, 
 * if the compression runs sucessfull
 */
bool decompress_image_file(const char* image_fname) {
	/// initialize variables
	INTEGER uncompressed_size = 0, compressed_size = 0;
	void* uncompressed_mem = NULL;
	void* compressed_mem = NULL;
	
	// Open file
	image_file = fopen(image_fname, "rb");
	if (!image_file)
		return false;
	
	ImageCompressionHeader ich;
	if (!READ_HEADER(&ich, sizeof(ich), image_file))
		return false;
	
	// get some space
	uncompressed_size = (ich.orginal_size * 1.01) + 12;
	uncompressed_mem = malloc(uncompressed_size);
	
	compressed_size = ich.compressed_size;
	compressed_mem = malloc(compressed_size);
	
	if (!compressed_mem || !uncompressed_mem) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	// read compressed data
	if (1 != fread(compressed_mem, compressed_size, 1, image_file)) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	// uncompress it
	if (!decompress_data(compressed_size,compressed_mem, &uncompressed_size, &uncompressed_mem)) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	// close file
	if (0 != fclose(image_file)) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	// Open file
	image_file = fopen(image_fname, "wb+");
	if (!image_file) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	// write new data into file
	if (   1 != fwrite(&image_header, sizeof(ImageHeader), 1, image_file)
		|| 1 != fwrite(uncompressed_mem, uncompressed_size, 1, image_file)) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	// close file
	if (0 != fclose(image_file)) {
		free(compressed_mem);
		free(uncompressed_mem);
		return false;
	}
	
	free(compressed_mem);
	free(uncompressed_mem);
	return true;
}
/// compressed the bytestream into a new one
bool decompress_data(const INTEGER size, const void* byte_sequence, 
		INTEGER* new_size, void** new_byte_sequence) {
	
	// initialize zlib struct
	z_stream zst;
	zst.zalloc = Z_NULL;
    zst.zfree = Z_NULL;
    zst.opaque = Z_NULL;
    
	// initialize return values
	void* ret_data = NULL;
    
    zst.avail_in = size;
    void* bs = (void*) byte_sequence;
    zst.next_in  = (Bytef*)bs;

    if (Z_OK != inflateInit(&zst))
        return false;
	
	zst.avail_out = *new_size;
	zst.next_out = (Bytef*)*new_byte_sequence;
	fprintf(stderr, "first byte: %x\n", zst.next_out);
	
	int ret = inflate(&zst, Z_FINISH);
	
	if (Z_STREAM_ERROR == ret || Z_STREAM_END != ret)
		return false;
	
	int temp = size - zst.avail_out;

	// end decompression
	if (Z_OK != inflateEnd(&zst))
		return false;
	
	// everything works fine
	*new_size = temp;
	*new_byte_sequence = ret_data;
	return true;
}

/** 
 * \bug currently not working
 * compresses the given byte_sequence and returns it by conpressed data, 
 * if the compression runs sucessfull
 */
bool compress_image_file(const char* image_fname) {
	/// initialize variables
	INTEGER uncompressed_size = 0, compressed_size = 0;
	void* uncompressed_mem;
	void* compressed_mem;
	
	// Get size of uncompressed data in file and memory for it.
	struct stat buf;
	stat(image_fname, &buf);
	uncompressed_size = buf.st_size - sizeof(ImageHeader);
	uncompressed_mem = malloc(uncompressed_size);
	if (!uncompressed_mem)
		return false;
	
	// Open file
	image_file = fopen(image_fname, "rb");
	if (!image_file) {
		free(uncompressed_mem);
		return false;
	}
	
	if (!READ_HEADER(uncompressed_mem, uncompressed_size, image_file)) {
		free(uncompressed_mem);
		return false;
	}
	
	if (!compress_data(uncompressed_size,uncompressed_mem, &compressed_size, &compressed_mem)) {
		free(uncompressed_mem);
		return false;
	}
	
	// close file
	if (0 != fclose(image_file)) {
		free(uncompressed_mem);
		return false;
	}
	
	// Open file
	image_file = fopen(image_fname, "wb+");
	if (!image_file) {
		free(uncompressed_mem);
		return false;
	}		
	
	ImageCompressionHeader ich;
	ich.compressed_size = compressed_size;
	ich.orginal_size    = uncompressed_size;
	
	// add compression information to image header
	SET_FORMAT_BIT(image_header.format_info, FORMAT_COMPRESSION_BIT);
	
	// write new data into file
	if (   1 != fwrite(&image_header,  sizeof(ImageHeader),  1, image_file)
		|| 1 != fwrite(&ich, sizeof(ImageCompressionHeader), 1, image_file)
		|| 1 != fwrite(compressed_mem, compressed_size,      1, image_file)) {
		free(uncompressed_mem);
		return false;
	}
	
	// close file
	if (0 != fclose(image_file)) {
		free(uncompressed_mem);
		return false;
	}
	
	free(uncompressed_mem);
	return true;
}

/// compresses the given data
bool compress_data(const INTEGER uncompressed_size, const void* uncompressed_mem,
		INTEGER* compressed_size, void** compressed_mem) {
	
	// compressionLevel in [0,9]	
	int compression_level = 9;
	
	// initialize zlib struct
	z_stream zst;
	zst.zalloc = Z_NULL;
	zst.zfree  = Z_NULL; 
	zst.opaque = Z_NULL;
	
	if (Z_OK != deflateInit(&zst, compression_level))
			return false;
	
	zst.avail_in  = uncompressed_size;
	zst.next_in   = (Bytef*) uncompressed_mem;
	zst.next_out  = NULL;

	// some magic compression stuff:
	// "[...]avail_out must be at least 0.1% larger than avail_in plus 12 bytes[...]"
	// (zlib documentation)
	INTEGER avail_out = (uncompressed_size * 1.01) + 12;
	zst.avail_out = avail_out;
	if (!zst.next_out)
		zst.next_out = (Bytef*)malloc(zst.avail_out);
	
	int ret = deflate(&zst, Z_FINISH);
	
	if (Z_STREAM_ERROR == ret || Z_STREAM_END != ret) {
		deflateEnd(&zst);
		return false;
	}

	// everything works fine
	*compressed_size = avail_out - zst.avail_out;
	*compressed_mem  = zst.next_out;
	fprintf(stderr, "first byte: %x\n", zst.next_out);
	
	if (Z_OK != deflateEnd(&zst))
		return false;

	return true;
}

/// \brief [en|de]crypt image with the given key and xor
bool _crypt_xor(INTEGER data_size, void** data, int32_t key) {
	
	// if this is a 64bit vm, then the key is duplicated
	INTEGER calculated_key = key;
	if (8 == sizeof(INTEGER))
		calculated_key = (calculated_key << 32) + key;
		
	for(INTEGER* curData = (INTEGER*)*data; curData < (INTEGER*)((INTEGER)*data + data_size); curData++)
		*curData = *curData ^ calculated_key;
	
	return true;
}

/// \brief decrypts the given image file
bool decrypt_image_file(const char* image_fname, const void* key) {
	debug_trace("Decrypting using key %i\n", *(int*)key);
	// Get size of uncompressed data in file and memory for it.
	struct stat buf;
	stat(image_fname, &buf);
	INTEGER size = buf.st_size - (sizeof(ImageHeader) + sizeof(ImageEncryptionHeader));
	void* mem = malloc(size);
	
	if (!mem)
		return false;
	
	// Open file
	image_file = fopen(image_fname, "rb");
	if (!image_file)
		return false;
	
	// read the ImageEncryptionHeader to specify the encryptionMethod
	ImageEncryptionHeader ieh;
	if (!READ_HEADER(&ieh, sizeof(ImageEncryptionHeader), image_file))
		return false;
	
	// reads the rest of the image file into the buffer for decryption
	if (1 != fread(mem, size, 1, image_file))
		return false;
	
	/// decrypt here
	switch (ieh.encryption_method) {
		case IMAGE_CRYPT_NONE:
			break;
		case IMAGE_CRYPT_XOR:
			if (!_crypt_xor(size, &mem, *((int32_t*)key)))
				return false;
			break;
		case IMAGE_CRYPT_RSA:
			break;
	}
	
	// close file
	if (0 != fclose(image_file))
		return false;
	
	// Open file
	image_file = fopen(image_fname, "wb+");
	if (!image_file)
		return false;
	
	// write new data into file
	if (   1 != fwrite(&image_header,  sizeof(ImageHeader), 1, image_file)
		|| 1 != fwrite(mem, size,    						1, image_file))
		return false;
	
	// close file
	if (0 != fclose(image_file))
		return false;
	
	return true;
}

/// \brief encrypts the given image_file
bool encrypt_image_file(const char* image_fname, int32_t encryption_method, const void* key) {
	debug_trace("Encrypting using key %i\n", *(int*)key);
	// Get size of uncompressed data in file and memory for it.
	struct stat buf;
	stat(image_fname, &buf);
	INTEGER size = buf.st_size - sizeof(ImageHeader);
	void* mem = malloc(size);
	
	if (!mem)
		return false;
	
	// Open file
	image_file = fopen(image_fname, "rb");
	if (!image_file)
		return false;
	
	if (!READ_HEADER(mem, size, image_file))
		return false;

	/// encrypt here
	switch (encryption_method) {
		case IMAGE_CRYPT_NONE:
			break;
		case IMAGE_CRYPT_XOR:
			if (!_crypt_xor(size, &mem, *((int32_t*)key)))
				return false;
			break;
		case IMAGE_CRYPT_RSA:
			break;
	}
	
	// close file
	if (0 != fclose(image_file))
		return false;
	
	// Open file
	image_file = fopen(image_fname, "wb+");
	if (!image_file)
		return false;
	
	// create ImageEncryptionHeader
	ImageEncryptionHeader ieh;
	ieh.encryption_method = encryption_method;
	
	// add encryption information to image header
	SET_FORMAT_BIT(image_header.format_info, FORMAT_ENCRYPTION_BIT);
	
	// write new data into file
	if (   1 != fwrite(&image_header,  sizeof(ImageHeader), 1, image_file)
		|| 1 != fwrite(&ieh, sizeof(ImageEncryptionHeader), 1, image_file)
		|| 1 != fwrite(mem,  size,							1, image_file) )
		return false;
	
	// close file
	if (0 != fclose(image_file))
		return false;

	return true;
}


/// copies source to destination
bool fcopy(const char *source, const char *dest) {
	int buffer_size = 4096;
	
	// open source file
	FILE* sf = fopen(source, "rb");
	if(!sf) return false;

	// open destination file
	FILE* df = fopen(dest, "wb+");
	if(!df) {
		fclose(sf);
		return false;
	}
	
	// try to allocate buffer
	void* buffer = malloc(buffer_size);
	if(!buffer) {
		fclose(sf); fclose(df);
		return false;
	}
	
	// copy all the data
	int bytes_read;
	do {
		bytes_read = fread(buffer, sizeof(char), buffer_size, sf);
		fwrite(buffer, sizeof(char), bytes_read, df);
	} while(!feof(sf));
	
	// shutdown copy shop
	free(buffer); fclose(sf); fclose(df);
	
	return true;
}
