/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include <stddef.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
 
#include "Universe.h"
#include "Shell.h"

#include <memory/memory.h>

#include <misc/Hashmap.h>
#include <misc/List.h>

#include <vmobjects/VMObject.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMMethod.h>
#include <vmobjects/VMArray.h>
#include <vmobjects/VMInvokable.h>
#include <vmobjects/VMBlock.h>
#include <vmobjects/VMBigInteger.h>
#include <vmobjects/VMImage.h>
#include <vmobjects/VMInteger.h>
#include <vmobjects/VMDouble.h>
#include <vmobjects/VMString.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/Symboltable.h>

#include <compiler/SourcecodeCompiler.h>

#include <interpreter/bytecodes.h>
#include <interpreter/Interpreter.h>
#include <interpreter/Disassembler.h>

#include <misc/debug.h>

// Here we go:
// externally refenced variables:

/** 
 * \brief Struct which represents an universe global
 * This is requiered to store the globals
 */
typedef struct _UniverseGlobal {
	pVMObject object;
	pVMSymbol symbol;
} UniverseGlobal;

typedef struct _UniverseState {	
	short dump_bytecodes;
	short gc_verbosity;
	int32_t globals_count;
} UniverseState;

// declaring the globals
pVMObject nil_object;
pVMObject true_object;
pVMObject false_object;

pVMObject system_object;

pVMClass object_class;
pVMClass class_class;
pVMClass metaclass_class;

pVMClass nil_class;
pVMClass integer_class;
pVMClass biginteger_class;
pVMClass array_class;
pVMClass method_class;
pVMClass symbol_class;
pVMClass image_class;
pVMClass frame_class;
pVMClass primitive_class;
pVMClass string_class;
pVMClass system_class;
pVMClass block_class;
pVMClass double_class;


//
// values for debug output verbosity levels (controllable via the command line)
//


short dump_bytecodes;
short gc_verbosity;

// private helper functions
static int  setup_class_path(pString cp);
static int  add_class_path(const pString restrict cp);
static void print_usage_and_exit(char* executable);
static int  get_path_class_ext(pString **tokens, pString arg);

//file-local variables
static pString* class_path=NULL;
static int cp_count=0;
static pHashmap globals_dictionary=NULL;
static pList globals_dictionary_symbols=NULL;
static pString image_name = NULL;
static int image_encrypt_key = 0; 
static pList classes_with_primitives = NULL; ///< pointers to classes with loaded primitives

///////////////////////////////////
#pragma mark private hepler functions

static int get_path_class_ext(pString **tokens, pString arg) {   
#define EXT_TOKENS 2    
     
    // result String array
    *tokens = (pString *)internal_allocate(EXT_TOKENS * sizeof(pString));
    
    // index last path/file separator
    int fp_index = SEND(arg, lastIndexOfChar, *file_separator);
    // index of filename/suffix separator
    int ssep_index = SEND(arg, lastIndexOfChar, '.');
        
    if(fp_index < 0) { //no new path
        internal_free(*tokens);
        return ERR_FAIL;    
    } else //copy path token
        (*tokens)[0] = SEND(arg, substring, 0, fp_index - 1);
    
    // copy filename to 2nd slot
    // strip suffix if present
    ssep_index = ( (ssep_index >= 0)  && (ssep_index > fp_index)) ?
                 (ssep_index - 1):
                 SEND(arg, length);
    // copy substring
    (*tokens)[1] = SEND(arg, substring, fp_index + 1, ssep_index);
    
    return ERR_SUCCESS;
}


/*
 *  setup 'path_separator'-separated Classpathes
 *  from cp
 */
static int setup_class_path(const pString restrict cp) {
    /*
     * affected globals:
     * affected file globals: class_path, cp_count
     */
    
    size_t count;
    if ((class_path = SEND(cp, tokenize, &count, path_separator))) {
        cp_count = count;
        return ERR_SUCCESS;
    } else
        return ERR_NOMEM;
}


/**
 * add cp to the class_path vector
 */
static int add_class_path(const pString restrict cp) {
    /*
     * affected    globals:
     * affected file globals: class_path, cp_count
     */
    
    // resize class_path-vector
    class_path = (pString*)realloc(class_path, (cp_count + 1) * sizeof(pString));
    if(!class_path) 
        return ERR_NOMEM;
    // copy it and increment
    class_path[cp_count++] = cp;
    
    return ERR_SUCCESS;    
    
}


static void print_usage_and_exit(char* executable) {
    // print the usage
    fprintf(stderr, "Usage: %s [-options] [args...]\n\n", executable);
    fprintf(stderr, "where options include:\n");
    fprintf(stderr, "    -cp <directories separated by %s>\n", path_separator);
    fprintf(stderr, "        set search path for application classes\n");
    fprintf(stderr, "    -d  enable disassembling (twice for tracing)\n");
	fprintf(stderr, "    -g  enable garbage collection details:\n" \
	                "        1x - print statistics when VM shuts down\n" \
	                "        2x - print statistics upon each collection\n" \
	                "        3x - print statistics and dump heap upon each " \
	                "collection\n");
	fprintf(stderr, "    -Hx set the heap size to x MB (default: 1 MB)\n");
    fprintf(stderr, "    -l <image name> (<key>) Load the specified image.\n");
    fprintf(stderr, "    -h  show this help\n");
    // exit
    Universe_exit(ERR_SUCCESS);
}


pVMClass new_metaclass_class(void) {
    // Allocate the metaclass classes
    pVMClass result = VMClass_new();
    SEND((pVMObject)result, set_class, VMClass_new());
    
    // Setup the metaclass hierarchy
    pVMObject mclass = (pVMObject)SEND((pVMObject)result, get_class);
    SEND(mclass, set_class, result);
    
    // Return the freshly allocated metaclass class
    return result;
}


pVMClass new_system_class(void) {
    // Allocate the new system class
    pVMClass system_class = VMClass_new();
    
    // Setup the metaclass hierarchy
    SEND((pVMObject)system_class, set_class, VMClass_new());
    pVMObject mclass = (pVMObject)SEND((pVMObject)system_class, get_class);
    SEND(mclass, set_class, metaclass_class);
    
    // Return the freshly allocated system class
    return system_class;
}


void initialize_system_class(pVMClass system_class,
    pVMClass super_class, const char* name
) {
    pString s_name = String_new(name);
    
    // Initialize the superclass hierarchy
    if(super_class != NULL) {
        SEND(system_class, set_super_class, super_class);
        pVMClass sys_class_class = SEND((pVMObject)system_class, get_class);
        pVMClass super_class_class = SEND((pVMObject) super_class, get_class);
        SEND(sys_class_class, set_super_class, super_class_class);
    } else {
        pVMClass sys_class_class = SEND((pVMObject)system_class, get_class);
        SEND(sys_class_class, set_super_class, class_class);
    }

    pVMClass sys_class_class = SEND((pVMObject)system_class, get_class);

    // Initialize the array of instance fields
    SEND(system_class, set_instance_fields, VMArray_new(0));
    SEND(sys_class_class, set_instance_fields, VMArray_new(0));

    // Initialize the array of instance invokables
    SEND(system_class, set_instance_invokables, VMArray_new(0));
    SEND(sys_class_class, set_instance_invokables, VMArray_new(0));
    
    // Initialize the name of the system class
    SEND(system_class, set_name, Universe_symbol_for(s_name));
    pString class_class_name =String_new_from(s_name);
    SEND(class_class_name, concatChars, " class");
    SEND(sys_class_class, set_name, Universe_symbol_for(class_class_name));
    
    // Insert the system class into the dictionary of globals
    Universe_set_global(SEND(system_class, get_name), (pVMObject)system_class);
}


void setup_object_system() {
    /*
     * affected globals:
     * affected file globals: globals_dictionary
     */

	///////////////////////////////////
    //     allocate the nil object
    nil_object = VMObject_new();
    // allocate the Metaclass classes
    metaclass_class = new_metaclass_class(); 
    // allocate the rest of the system classes
    object_class    = new_system_class();
    nil_class       = new_system_class();
    class_class     = new_system_class();
    array_class     = new_system_class();
    symbol_class    = new_system_class();
    method_class    = new_system_class();
    image_class     = new_system_class();
    integer_class   = new_system_class();
    biginteger_class= new_system_class();
    frame_class     = new_system_class();
    primitive_class = new_system_class();
    string_class    = new_system_class();
    double_class    = new_system_class();
        
    // setup the class reference for the nil object
    SEND(nil_object, set_class, nil_class);
        
    // initialize the system classes.
    initialize_system_class(object_class, NULL, "Object");
    initialize_system_class(class_class, object_class, "Class");
    initialize_system_class(metaclass_class, class_class, "Metaclass");
    initialize_system_class(nil_class, object_class, "Nil");
    initialize_system_class(array_class, object_class, "Array");
    initialize_system_class(method_class, array_class, "Method");
    initialize_system_class(symbol_class, object_class, "Symbol");
    initialize_system_class(image_class, object_class, "Image");
    initialize_system_class(integer_class, object_class, "Integer");
    initialize_system_class(biginteger_class, object_class,
                                     "BigInteger");
    initialize_system_class(frame_class, array_class, "Frame");
    initialize_system_class(primitive_class, object_class,
                                     "Primitive");
    initialize_system_class(string_class, object_class, "String");
    initialize_system_class(double_class, object_class, "Double");
    
    // load methods and fields into the system classes
    Universe_load_system_class(object_class);
    Universe_load_system_class(class_class);
    Universe_load_system_class(metaclass_class);
    Universe_load_system_class(nil_class);
    Universe_load_system_class(array_class);
    Universe_load_system_class(method_class);
    Universe_load_system_class(symbol_class);
    Universe_load_system_class(image_class);
    Universe_load_system_class(integer_class);
    Universe_load_system_class(biginteger_class);
    Universe_load_system_class(frame_class);
    Universe_load_system_class(primitive_class);
    Universe_load_system_class(string_class);
    Universe_load_system_class(double_class);

    // load the generic block class
    block_class = Universe_load_class(Universe_symbol_for_chars("Block"));
    
    // setup the true and false objects
    true_object = VMObject_new_instance(Universe_load_class(Universe_symbol_for_chars("True")));
    false_object = VMObject_new_instance(Universe_load_class(Universe_symbol_for_chars("False")));
    
    // load the system class and create an instance of it
    system_class = Universe_load_class(Universe_symbol_for_chars("System"));
    pVMObject system_object = VMObject_new_instance(system_class);
    
    // put special objects and classes into the dictionary of globals
    Universe_set_global(Universe_symbol_for_chars("nil"), nil_object);
    Universe_set_global(Universe_symbol_for_chars("true"), true_object);
    Universe_set_global(Universe_symbol_for_chars("false"), false_object);
    Universe_set_global(Universe_symbol_for_chars("system"), system_object);
    Universe_set_global(Universe_symbol_for_chars("System"), (pVMObject)system_class);
    Universe_set_global(Universe_symbol_for_chars("Block"), (pVMObject)block_class);
}

/**
 * \brief Prepares Universe to execute an application.
 * 
 * Readies Universe for execution of an application, creating a
 * bootstrap-frame and making system object and command-line arguments available.
 */
void prepare_for_application(int argc, pString* argv, pVMMethod bootstrap_method) {
    /* only trace bootstrap if the number of cmd-line "-d"s is > 2 */
    int trace = 2 - dump_bytecodes;
    if(!(trace > 0)) dump_bytecodes = 1;
    
    // convert the arguments into an array
    pVMArray arguments_array = VMArray_new_from_strings(argc, argv);
        
    // create a fake bootstrap frame with the system object on the stack
    pVMFrame bootstrap_frame;
    Interpreter_push_new_frame(bootstrap_method, &bootstrap_frame);

    pVMObject system_object = Universe_get_global(Universe_symbol_for_chars("system"));

    SEND(bootstrap_frame, push, system_object);
    SEND(bootstrap_frame, push, (pVMObject)arguments_array);
            
    // lookup the initialize invokable on the system class
    pVMObject initialize =
        (pVMObject)SEND(system_class, lookup_invokable,
                        Universe_symbol_for_chars("initialize:"));
        
    // invoke the initialize invokable
    TSEND(VMInvokable, initialize, invoke, bootstrap_frame);

    // reset "-d" indicator
    if(!(trace>0)) dump_bytecodes = 2 - trace;
}


////////////////


#pragma mark Extern Callable Functions

void Universe_exit(int err) {
    if(gc_verbosity > 0)
        gc_stat();
    Universe_destruct();
    exit(err);
}


void Universe_error_exit(const char* restrict err) {
    debug_error("Runtime error: %s\n", err);
    Universe_exit(ERR_FAIL);
}


pString* Universe_handle_arguments(int* vm_argc, int argc, char** argv) {
    //VM Arguments to return
    pString* vm_args = (pString*)internal_allocate(sizeof(pString*));
    *vm_argc = 0;
    
    dump_bytecodes = 0;
    gc_verbosity = 0;
    
    /* iterate agruments (argv[>0])
     *
     */
    for (int i = 1; i < argc ; i++) {
        if (strcmp(argv[i], "-cp") == 0) { // Classpath
            if((argc == i+1) // no CP given
                 || class_path // CP already set up.
                 ) 
                print_usage_and_exit(argv[0]);
            
            // setup & skip class path
            setup_class_path(String_new(argv[++i]));

        } else if(strcmp(argv[i], "-l") == 0) { // Classpath
                if((argc == i+1)) { // no image name given
                	fprintf(stderr, "No image name given.\n");
                    print_usage_and_exit(argv[0]);
                }
                
                fprintf(stderr, "Trying to load '%s'\n", argv[ i+1 ]);
                image_name = String_new(argv[ i+1 ]);
                
                // get key, if possible
                if (argc > i+2) {
                	image_encrypt_key = atoi(argv[ i+2 ]);
                }

        } else if(strcmp(argv[i], "-d") == 0) { // Dump  bytecode
            dump_bytecodes++;
        } else if(strcmp(argv[i], "-g") == 0) {
            gc_verbosity++;
        } else if(argv[i][0] == '-' && argv[i][1] == 'H') {
            int heap_size = atoi(argv[i] + 2);
            gc_set_heap_size(heap_size);
        } else if((strcmp(argv[i], "-h") == 0) ||
            (strcmp(argv[i], "--help") == 0)
        ) {
            print_usage_and_exit(argv[0]);
        } else {
            /* copy remaining args for vm */
            
            // out of memory for the vm_args pointer
            if(!vm_args) Universe_exit(ERR_NOMEM);
                        
            // temp vector for path
            pString* ext_path_tokens = NULL;
            pString tmp_string = String_new(argv[i]);
            if(get_path_class_ext(&ext_path_tokens, tmp_string) ==
               ERR_SUCCESS
            ) {
                // got additional path, so add it
                add_class_path(ext_path_tokens[0]);
                                
                // copy filename to args
                /// TODO!!: resizing!!
                vm_args[(*vm_argc)++] = ext_path_tokens[1];

                // guaranteed not to be used/referenced any more
                memory_free(ext_path_tokens);
                
            } else {
                // no path, so copy arg only    
                // copy filename to args
                vm_args[(*vm_argc)++] = String_new_from(tmp_string);
            }
            // will not be referenced, because copied
            SEND(tmp_string, free);
        }
    }

    // Add local dir to class_path
    add_class_path(String_new("."));
    
    return vm_args;
}


void Universe_initialize(int argc, pString* argv) {
    /*
     * affected globals:
     * affected file globals: globals_dictionary
     */

    // setup the Hashmap for all globals
    globals_dictionary = Hashmap_new();
    
    // setup the hashmap for the keys of globals_dictionary
    globals_dictionary_symbols = List_new();
    
    // init the Symboltable
    Symbol_table_init();
    // classes with loaded primitives
    classes_with_primitives = List_new();
    
    // initialize the system
    // this is either achieved by loading an image or by intializing a new one
    bool initialize_new = false;
    if (image_name) {
    	if (!image_load_with_name(image_name, image_encrypt_key)) {
    		Universe_error_exit("Image loading failed, exiting.\n");
    		initialize_new = true;
    	}
    	else
    		fprintf(stderr, "Image loading succeeded.\n");
    }
    else initialize_new = true; 
    
    if (initialize_new) {
    	setup_object_system();
    	
        // create a fake bootstrap method to simplify later frame traversal
        pVMMethod bootstrap_method = Interpreter_create_bootstrap_method(system_class);     

        // start the shell if no filename is given
        if(argc == 0) {
          Shell_set_bootstrap_method(bootstrap_method);
          Shell_start();
          return;
        }
        
        prepare_for_application(argc, argv, bootstrap_method);
    }
        
    // start the interpreter
    Interpreter_start();
}


/**
 * Universe Destructor
 * Free "member" variables
 */    
void Universe_destruct(void) {
    /*
     * affected    globals:
     * affected file globals: class_path, globals_dictionary
     */

    Symbol_table_destruct();
    if(globals_dictionary)
        SEND(globals_dictionary, free);
    FREE_ARRAY(class_path, cp_count);
}


void Universe_assert(bool value) {
    if(!value)
      // for now we just print something whenever an assertion fails
      debug_error("Assertion failed!\n");
}


pVMSymbol Universe_symbol_for(pString string) {
    // Lookup the symbol in the symbol table
    pVMSymbol result = Symbol_table_lookup(SEND(string, chars));
    
    // return found or newly created symbol
    return result ?
           result : // found
           VMSymbol_new_from_string(string); //new
}


pVMSymbol Universe_symbol_for_chars(const char* restrict string) {
    // Lookup the symbol in the symbol table
    pVMSymbol result = Symbol_table_lookup(string);
    
    // return found or newly created symbol
    return result ? result : VMSymbol_new_from_string(String_new(string)); //new
}


pVMImage Universe_new_image(pVMInteger file_size, pVMString file_name, ImageHeader image_header, ImageInfoHeader info_header) {
    // Allocate a new integer and set its class to be the integer class
    pVMImage result = VMImage_new_with(file_size, file_name, image_header, info_header);
    SEND((pVMObject)result, set_class, image_class);
    
    // Return the freshly allocated integer
    return result;
}


pHashmap Universe_get_globals_dictionary() {
    return globals_dictionary;
}


pVMObject Universe_get_global(pVMSymbol name) {
    // Return the global with the given name if it's in the dictionary of
    // globals
    if(Universe_has_global(name))
        return (pVMObject)SEND(globals_dictionary, get, name);
    
    // Global not found
    return NULL;
}


void Universe_set_global(pVMSymbol name, pVMObject value) {
    // Insert the given value into the dictionary of globals
    SEND(globals_dictionary, put, name, value);
    // register key for symbol
    pList pl = globals_dictionary_symbols;
    SEND(globals_dictionary_symbols, addIfAbsent, name);
}


bool Universe_has_global(pVMSymbol name) {
    // Returns if the Universe has a value for the global of the given name
    return SEND(globals_dictionary, contains_key, name);
}


pVMClass Universe_get_block_class(void) {
    return block_class;
}


pVMClass Universe_get_block_class_with_args(size_t number_of_arguments) {
    // Compute the name of the block class with the given number of arguments
    char block_name[7];
    Universe_assert(number_of_arguments <10); // buffer overflow otherwise
    sprintf(block_name, "Block%d", number_of_arguments);
    pVMSymbol name = Universe_symbol_for_chars(block_name);
    
    // Lookup the specific block class in the dictionary of globals and return
    // it
    if(Universe_has_global(name)) 
        return (pVMClass)Universe_get_global(name);

    // Get the block class for blocks with the given number of arguments
    pVMClass result = Universe_load_class_basic(name, NULL);
    
    // Add the appropriate value primitive to the block class
    SEND(result, add_instance_primitive,
         VMBlock_get_evaluation_primitive(number_of_arguments));
    
    // Insert the block class into the dictionary of globals
    Universe_set_global(name, (pVMObject)result);
    
    // Return the loaded block class
    return result;
}


pVMClass Universe_load_class(pVMSymbol name) {
    // Check if the requested class is already in the dictionary of globals
    if(Universe_has_global(name)) 
        return (pVMClass)Universe_get_global(name);
    
    // Load the class
    pVMClass result = Universe_load_class_basic(name, NULL);
        
    // check class loading.
    if(!result) {
        pString name_s = SEND(name, get_string);
        debug_error("can't load class:\t'%s'\n", SEND(name_s, chars));
        Universe_exit(ERR_FAIL);
    }

    // Load primitives (if necessary) and return the resulting class
    if(SEND(result, has_primitives) || SEND(result->clazz, has_primitives)) 
        SEND(result, load_primitives, class_path, cp_count);
    return result;
}


void Universe_load_system_class(pVMClass system_class) {
    // Load the system class
    pVMClass result =
        Universe_load_class_basic(SEND(system_class, get_name), system_class);
    
    // check class loading.
    if(!result) {
        pVMSymbol cname = SEND(system_class, get_name);
        debug_error("can't load system class:\t%s", SEND(cname, get_string));
        Universe_exit(ERR_FAIL);
    }

    // Load primitives (if necessary) 
    if(SEND(result, has_primitives) || SEND(result->clazz, has_primitives)) 
        SEND(result, load_primitives, class_path, cp_count);
}


pVMClass Universe_load_class_basic(pVMSymbol name, pVMClass system_class) {
    pString s_name = SEND(name, get_string);
    debug_log("%s\n", SEND(s_name, chars));

    pVMClass result;
    // Try loading the class from all different paths
    for(int i = 0; i < cp_count; i++) {
        // Load the class from a file and return the loaded class
        result = SourcecodeCompiler_compile_class(class_path[i], 
                                                  SEND(name, get_string),
                                                  system_class);
        if(result) {
            if(dump_bytecodes) {
                Disassembler_dump(SEND(result, get_class));
                Disassembler_dump(result);
            }
            // found class in class_path, so return it
            return result; 
        }
    }
    // The class could not be found.
    return NULL;
}


pVMClass Universe_load_shell_class(pString stmt) {
    // Load the class from a stream and return the loaded class
    pVMClass result = SourcecodeCompiler_compile_class_string(stmt, NULL);
    if(dump_bytecodes)
        Disassembler_dump(result);
    return result;    
}

bool Universe_is_global(const pVMObject subject) {
	return SEND(globals_dictionary, contains, subject);
}

bool Universe_is_system_class(const pVMClass subject) {
	return (
			   (subject == object_class)
			|| (subject == class_class)
			|| (subject == metaclass_class)

			|| (subject == nil_class)
			|| (subject == integer_class)
			|| (subject == biginteger_class)
			|| (subject == array_class)
			|| (subject == method_class)
			|| (subject == symbol_class)
			|| (subject == frame_class)
			|| (subject == primitive_class)
			|| (subject == string_class)
			|| (subject == system_class)
			|| (subject == block_class)
			|| (subject == double_class)
			);
}

/// test function for heap reallocation during runtime
void Universe_update_references(const INTEGER difference) {
	nil_object = (pVMObject)((INTEGER)nil_object + difference);
	true_object = (pVMObject)((INTEGER)true_object + difference);
	false_object = (pVMObject)((INTEGER)false_object + difference);
	
	//system_object = (pVMObject)((INTEGER)system_object + difference);
	
	object_class = (pVMClass)((INTEGER)object_class + difference);
	class_class = (pVMClass)((INTEGER)class_class + difference);
	metaclass_class = (pVMClass)((INTEGER)metaclass_class + difference);
	
	nil_class = (pVMClass)((INTEGER)nil_class + difference);
	integer_class = (pVMClass)((INTEGER)integer_class + difference);
	biginteger_class = (pVMClass)((INTEGER)biginteger_class + difference);
	array_class = (pVMClass)((INTEGER)array_class + difference);
	method_class = (pVMClass)((INTEGER)method_class + difference);
	symbol_class = (pVMClass)((INTEGER)symbol_class + difference);
	frame_class = (pVMClass)((INTEGER)frame_class + difference);
	primitive_class = (pVMClass)((INTEGER)primitive_class + difference);
	string_class = (pVMClass)((INTEGER)string_class + difference);
	system_class = (pVMClass)((INTEGER)system_class + difference);
	block_class = (pVMClass)((INTEGER)block_class + difference);
	double_class = (pVMClass)((INTEGER)double_class + difference);
		
    SEND(globals_dictionary, update_references, difference);
}

/// \todo document me 
static UniverseState uniState;

bool Universe_serialize(void) {
	// store the Universe
	uniState.globals_count = SEND(globals_dictionary_symbols, size); 
	uniState.dump_bytecodes		= dump_bytecodes;
	uniState.gc_verbosity		= gc_verbosity; 
		
	// store the universe state
	if (!image_store_byte_sequence(sizeof(uniState), &uniState))
		return false;
	
	// save pointers to classes containing symbols and globals of the universe
	UniverseGlobal globals[uniState.globals_count];
	for (int i = 0; i < uniState.globals_count; i++) {
		globals[i].symbol = (pVMSymbol)SEND(globals_dictionary_symbols, get, i);
		globals[i].object = Universe_get_global(globals[i].symbol);
	}
	
	if (!image_store_byte_sequence(sizeof(UniverseGlobal) * uniState.globals_count, &(globals[0])))
		return false;
		
	
	// save pointers to classes containing loaded primitives
	int number = SEND(classes_with_primitives, size);
	pVMClass pointers[number];
	for (int i = 0; i < number; i++)
		pointers[i] = (pVMClass)SEND(classes_with_primitives, get, i);
	
	if (!image_store_byte_sequence(sizeof(pVMClass) * number, &(pointers[0])))
		return false;

	return true;
}

bool Universe_deserialize(INTEGER difference) {
	INTEGER data_size;
	UniverseState* tempState = (UniverseState*)image_load_byte_sequence(&data_size, false);

	if (!tempState)
		return false;
	
	if (sizeof(*tempState) != data_size)
		return false;
	
	uniState = *tempState;
	
	dump_bytecodes = uniState.dump_bytecodes;
	gc_verbosity   = uniState.gc_verbosity;


	// reinitialize lists
    SEND(globals_dictionary, free);
    SEND(globals_dictionary_symbols, free);
    Symbol_table_destruct();
    SEND(classes_with_primitives, free);
    
    globals_dictionary = Hashmap_new();
    globals_dictionary_symbols = List_new();
    Symbol_table_init();
    classes_with_primitives = List_new();
    
    
	// reassembling the globals	
	INTEGER bytes_read;
	UniverseGlobal* globals = (UniverseGlobal*)image_load_byte_sequence(&bytes_read, false);
	
	if (!globals)
		return false;
	
	int number = bytes_read / sizeof(UniverseGlobal);
	
	if (number != uniState.globals_count)
		return false;
	
	UniverseGlobal global;
	for (int i = 0; i < number; i++) {
		global.symbol = (pVMSymbol)((INTEGER)globals[i].symbol + difference);
		global.object = (pVMObject)((INTEGER)globals[i].object + difference);		
		update_object_references((pVMObject)global.object, difference);
		update_object_references((pVMObject)global.symbol, difference);		
		Universe_set_global(global.symbol, global.object);
	}
	
	// load pointers to classes containing primitives
	pVMClass* classes = (pVMClass*)image_load_byte_sequence(&bytes_read, false);
	
	if (!classes)
		return false;
	
	number = bytes_read / sizeof(pVMClass);
	pVMClass clazz;
	for (size_t i = 0; i < number; i++) {
		clazz = (pVMClass)((INTEGER)classes[i] + difference);
		Universe_register_class_with_primitives(clazz);
	}
	
	number = SEND(classes_with_primitives, size);
	for (size_t i = 0; i < number; i++) {
		clazz = (pVMClass)SEND(classes_with_primitives, get, i);
		update_object_references((pVMObject)clazz, difference);
	}


	return true;	
}

// Serializes non-heap data of symbols and globals
bool Universe_serialize_object_data(void) {
	int number = SEND(globals_dictionary_symbols, size);
	pVMSymbol symbol;
	for (size_t i = 0; i < number; i++) {
		symbol = (pVMSymbol)SEND(globals_dictionary_symbols, get, i);
		serialize_object_data((pVMObject)symbol);
		serialize_object_data((pVMObject)Universe_get_global(symbol));
	}
	
	number = SEND(classes_with_primitives, size);
	pVMClass clazz;
	for (size_t i = 0; i < number; i++) {
		clazz = (pVMClass)SEND(classes_with_primitives, get, i);
		serialize_object_data((pVMObject)clazz);
	}

	return true;
}

// Restores non-heap data from the image file, i.e. strings.
bool Universe_deserialize_object_data(void) {
	size_t number = SEND(globals_dictionary_symbols, size);
	pVMSymbol symbol;
	for (size_t i = 0; i < number; i++) {
		symbol = (pVMSymbol)SEND(globals_dictionary_symbols, get, i);
		deserialize_object_data((pVMObject)symbol);
		deserialize_object_data((pVMObject)Universe_get_global(symbol));
		Symbol_table_insert(symbol);
	}
	
	number = SEND(classes_with_primitives, size);
	pVMClass clazz;
	for (size_t i = 0; i < number; i++) {
		clazz = (pVMClass)SEND(classes_with_primitives, get, i);
		deserialize_object_data((pVMObject)clazz);
	}
	
	return true;
}

// saves a reference to all classes that loaded primitives
void Universe_register_class_with_primitives(const pVMClass clazz) {
	SEND(classes_with_primitives, addIfAbsent, clazz);
}

// Restores the list of classes with primitives and loads their code dynamically. 
bool Universe_finalize(INTEGER difference) {
	nil_object = Universe_get_global(Universe_symbol_for_chars("nil"));
	true_object = Universe_get_global(Universe_symbol_for_chars("true"));
	false_object = Universe_get_global(Universe_symbol_for_chars("false"));
	
	system_object = Universe_get_global(Universe_symbol_for_chars("system"));
	
	object_class = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Object"));
	class_class = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Class"));
	metaclass_class = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Metaclass"));
	
	nil_class        = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Nil"));
	integer_class    = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Integer"));
	biginteger_class = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Biginteger"));
	array_class      = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Array"));
	method_class     = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Method"));
	symbol_class     = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Symbol"));
	frame_class      = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Frame"));
	primitive_class  = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Primitive"));
	string_class     = (pVMClass)Universe_get_global(Universe_symbol_for_chars("String"));
	system_class     = (pVMClass)Universe_get_global(Universe_symbol_for_chars("System"));
	block_class      = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Block"));
	double_class     = (pVMClass)Universe_get_global(Universe_symbol_for_chars("Double"));
	
	int number = SEND(classes_with_primitives, size);
	pVMClass clazz;
	for (size_t i = 0; i < number; i++) {
		clazz = (pVMClass)SEND(classes_with_primitives, get, i);
		SEND(clazz, load_primitives, class_path, cp_count);
	}
	return true;
}
