/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include "VMInteger.h"
#include "VMTaggedIntegerUtils.h"

#include <memory/gc.h>

// forward declarations
int printf(const char * restrict format, ...);
void Universe_exit(int);
void Universe_error_exit(const char* restrict);
int _VMObject_get_field_index(void* _self, pVMSymbol name);
pVMSymbol _VMObject_get_field_name(void* _self, int index);
void _VMObject_send(void* _self, pString selector_string, pVMObject* arguments, int argc);
extern pVMClass integer_class;
// end forward declarations


struct _VMInteger _VMInteger_Global_Box_Instance;
pVMInteger _VMInteger_Global_Box = 0;
pVMInteger VMInteger_Global_Box(void) {
    if(_VMInteger_Global_Box == 0) {
        _VMInteger_Global_Box = (pVMInteger)&_VMInteger_Global_Box_Instance;
        _VMInteger_Global_Box->_vtable     = VMInteger_vtable();
        _VMInteger_Global_Box->clazz       = integer_class; 
    }
    return _VMInteger_Global_Box;
}


//
//  Class Methods (Starting with VMInteger_) 
//


// overwritten parent methods
void _VMInteger_free(void* _self) {
    // nop
}

int32_t _VMInteger_object_size(void* _self) {
	return (VMINTEGER_IS_TAGGED(_self)) ? sizeof(void*) : sizeof(VMInteger);
}

pVMClass _VMInteger_get_class(void* _self) {
    return integer_class;
}

void _VMInteger_set_class(void* _self, pVMClass value) {
    // nop
}

pVMObject _VMInteger_get_field(void* _self, int index) {
    return VMInteger_Global_Box()->fields[index];
}

void _VMInteger_set_field(void* _self, int index, pVMObject value) {
    VMInteger_Global_Box()->fields[index] = value;
}

int _VMInteger_get_number_of_fields(void* _self) {
    return 1;
}


/** 
 * Create a new VMInteger
 */
pVMInteger VMInteger_new(void) {
    return VMInteger_new_with(0);
}


/** 
 * Create a new VMInteger with initial value
 */
pVMInteger VMInteger_new_with(const int32_t integer) {
	if(integer < VMTAGGEDINTEGER_MIN || integer > VMTAGGEDINTEGER_MAX) {
		pVMInteger result = (pVMInteger)gc_allocate_object(sizeof(VMInteger));
		if(result) {
			((pVMInteger)DEREF(result))->_vtable = VMInteger_vtable();
			INIT(result, integer);
		}
		return result;
	} else {
		return (pVMInteger)VMINTEGER_TAG(integer);
	}
}


/**
 * Initialize a VMInteger
 */
void _VMInteger_init(void* _self, ...) {
    pVMInteger self = (pVMInteger)DEREF(_self);
    SUPER(VMObject, _self, init, NUMBER_OF_OBJECT_FIELDS(VMObject));
    
    va_list args; va_start(args, _self);
    self->embedded_integer = va_arg(args, int32_t);
    self->hash = self->embedded_integer;
    va_end(args);
    
    SEND(self, set_class, integer_class);
}

//
//  Instance Methods (Starting with _VMInteger_) 
//
//
int32_t _VMInteger_get_embedded_integer(void* _self) {
	if (!VMINTEGER_IS_TAGGED(_self)) {
		pVMInteger self = (pVMInteger)DEREF(_self);
		return self->embedded_integer;
	} else {
		return (int32_t)VMINTEGER_UNTAG(_self);
	}
}


void _VMInteger_set_embedded_integer(void* _self, const int32_t embedded) {
	if (!VMINTEGER_IS_TAGGED(_self)) {
		pVMInteger self = (pVMInteger)DEREF(_self);
		self->embedded_integer = embedded;
		self->hash = self->embedded_integer;
	} else {
		Universe_error_exit(
		    "_VMInteger_set_embedded_integer not implemented\n");
	}
}


int32_t _VMInteger_hash_code(void* _self) {
    if(!VMINTEGER_IS_TAGGED(_self)) {
        return ((pVMObject)_self)->hash;
    } else {
        // return (int32_t)VMINTEGER_UNTAG(_self);
		return (int32_t)_self;
    }
}


//
// The VTABLE-function
//
//
static VTABLE(VMInteger) _VMInteger_vtable;
bool VMInteger_vtable_inited = false;

VTABLE(VMInteger)* VMInteger_vtable(void) {
    if(! VMInteger_vtable_inited) {
        *((VTABLE(VMObject)*)&_VMInteger_vtable) = *VMObject_vtable();
	
        // from ooobject
        _VMInteger_vtable.free        = METHOD(VMInteger, free);
        _VMInteger_vtable.object_size = METHOD(VMInteger, object_size);
        _VMInteger_vtable.hash_code   = METHOD(VMInteger, hash_code);

        // from vmobject
        _VMInteger_vtable.get_class = METHOD(VMInteger, get_class);
        _VMInteger_vtable.set_class = METHOD(VMInteger, set_class);
        _VMInteger_vtable.get_number_of_fields =
            METHOD(VMInteger, get_number_of_fields);
        _VMInteger_vtable.get_field = METHOD(VMInteger, get_field);
        _VMInteger_vtable.set_field = METHOD(VMInteger, set_field);

        _VMInteger_vtable.init = METHOD(VMInteger, init);
        _VMInteger_vtable.get_embedded_integer =
            METHOD(VMInteger, get_embedded_integer);
        _VMInteger_vtable.set_embedded_integer =
            METHOD(VMInteger, set_embedded_integer);
        
        VMInteger_vtable_inited = true;
    }
    return &_VMInteger_vtable;
}
