/*
 * $Id: gc.c 241 2008-04-24 09:00:06Z michael.haupt $
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
 */

#include <module.VM>
#include <module.Memory>

#include <vmobjects/objectformats.h>
#include <vmobjects/OOObject.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/VMObject.h>
#include <vmobjects/VMPrimitive.h>
#include <vmobjects/VMEvaluationPrimitive.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// count allocated objectes
int number_of_allocated_objects = 0;
// count live objects
int number_of_live_objects = 0;
// cound immortal objects
int number_of_immortal_objects = 0;

long number_of_inc = 0;
long number_of_dec = 0;

void* gc_allocate(size_t size) {
    if(size == 0)
        return NULL;
    void* result = calloc(1, size);
    //debug_trace("Allocated Mem at %p Size: %d\n", result, size);
    if(!result) {
        debug_error("Failed to allocate %d bytes. Panic.\n", (int)size);
        Universe_exit(-1);
    }
//#ifdef GC_DEBUG
    number_of_allocated_objects++;
    number_of_live_objects++;
    //fprintf(stderr,"[INFO] Allocated object %p (# %d) \n", result, number_of_live_objects);
//#endif    
    return result;
}


void* gc_allocate_object(size_t size) {
    void* o = gc_allocate(size);
    if(o)
        ((pOOObject)o)->object_size = size;
    return o;
}


char* gc_allocate_string(const char* restrict str) {
    return strdup(str);
}


void gc_free(void* ptr) {
//#ifdef GC_DEBUG
//    fprintf(stderr, "[INFO] Free object %p (%s)\n", ptr, TYPE_OF(ptr));
    number_of_live_objects--;
//#endif
    internal_free(ptr);
}


// increase reference counter of an object
void gc_inc_reference_counter(pVMObject object) {
    if (object == NULL) return;
	
    // if object is an global that should not be removed
    if (object->gc_field == -1) return;

    // if object is already marked for deletion there is a bug in the gc_inc/dec application
    if (object->gc_field == -2) Universe_error_exit("GC_INC/DEC are buggy, an object marked for deletion has got an inc");
    
    // increment reference counter
    object->gc_field++;
    number_of_inc++;
    
    #ifdef GC_DEBUG
    /***/
        fprintf(stderr, "[INFO] Incremented ref counter of object %p (%s)", object, TYPE_OF(object));
        /*pVMClass object_class = SEND((pVMObject)object, get_class);
        if (object_class != NULL && object_class != nil_object && object_class->gc_field) {
            pVMSymbol class_symbol = SEND(object_class, get_name);
            if (class_symbol != NULL && class_symbol != nil_object && class_symbol->gc_field) {
                pString class_name = SEND(class_symbol, get_string);
                if (class_name != NULL) {
                    fprintf(stderr, "class %s ", class_name->chars);
                }
            }
        }*/
        
        if (IS_A(object, String)) {
        	fprintf(stderr, "(string: %s)", ((pString)object)->chars);
        }
        fprintf(stderr, "to %d\n", object->gc_field);
    /***/
    #endif	
}

// decrease reference counter of an object
void gc_dec_reference_counter(pVMObject object) {
	if (object == NULL) return;
	
    // if object is an global or the delete process of this object is still in progress
    if (object->gc_field == -1 || object->gc_field == -2) return;
    
    // decrement reference counter of this object
    object->gc_field--;
    // this is done to find not correctly inc'ed objects
    if (object->gc_field == -1 || object->gc_field == 0) object->gc_field = -2;

    number_of_dec++;
    
    #ifdef GC_DEBUG
    /***/
        fprintf(stderr, "[INFO] Decremented ref counter of object %p (%s) ", object, TYPE_OF(object));
        /*pVMClass object_class = SEND((pVMObject)object, get_class);
        if (object_class != NULL && object_class != nil_object && object_class->gc_field) {
        	pVMSymbol class_symbol = SEND(object_class, get_name);
            if (class_symbol != NULL && class_symbol != nil_object && class_symbol->gc_field) {
                pString class_name = SEND(class_symbol, get_string);
                if (class_name != NULL) {
                    fprintf(stderr, "class %s ", class_name->chars);
                }
            }
        }*/
        if (IS_A(object, String)) {
            fprintf(stderr, "(string: %s)", ((pString)object)->chars);
        }
        fprintf(stderr, "to %d\n", object->gc_field);
    /***/
    #endif
    
    // this object is not referenced by any other object, so it can be deleted
    if (object->gc_field == -2) {
    	
    	for(int i = 0;i < SEND(object, get_number_of_fields); i++) {
    		// make sure not to dec on those embedded data for primitives, there aren't VMObjects 
    		if (!((i == 3  || i == 4) && (IS_A(object, VMPrimitive) || IS_A(object, VMEvaluationPrimitive)))) {
    			gc_dec_reference_counter(SEND(object, get_field, i));
    		}
        }
    	
        SEND(object, free);
        
        //#ifdef GC_DEBUG
        //fprintf(stderr, "[INFO] Deleted object %p (remaining: #%d) \n", object, number_of_live_objects);
        number_of_live_objects--;
        //#endif
    }
}

// set an object as global object, so it is not considered by reference counting
void gc_set_global_object(pVMObject object) {
    // set the reference counter to -1 indicates a global object
    object->gc_field = -1;
}

// returns the number of allocated objects
int gc_get_number_of_allocated_objects() {
    return number_of_allocated_objects;	
}

// return the number of allocated objects minus deleted objects
int gc_get_number_of_live_objects() {
    return number_of_live_objects;	
}

// return the number of immortal objects
int gc_get_number_of_immortal_objects() {
    return number_of_immortal_objects;
}

int allocated_objects_before = 0;
int live_objects_before = 0;
int immortal_objects_before = 0;

void gc_prepare_statistics(void) {
    allocated_objects_before = gc_get_number_of_allocated_objects();
    live_objects_before = gc_get_number_of_live_objects();
    immortal_objects_before = gc_get_number_of_immortal_objects();
}

void gc_statistics(void) {
    int allocated_objects_after = gc_get_number_of_allocated_objects();
    int live_objects_after = gc_get_number_of_live_objects();
    int immortal_objects_after = gc_get_number_of_immortal_objects();
    
    fprintf(stderr, "Allocated objects: before: %d, after: %d, diff: %d\n", 
        allocated_objects_before, allocated_objects_after,
        allocated_objects_after - allocated_objects_before);
    fprintf(stderr, "Immortal objects: before: %d, after: %d, diff: %d\n",
        immortal_objects_before, immortal_objects_after,
        immortal_objects_after - immortal_objects_before);
    fprintf(stderr, "Live objects: before: %d, after: %d, diff: %d\n", 
        live_objects_before, live_objects_after,
        live_objects_after - live_objects_before);
    fprintf(stderr, "Freed objects: before %d, after: %d, freed: %d\n",
        allocated_objects_before - live_objects_before,
        allocated_objects_after - live_objects_after,
        (allocated_objects_after-live_objects_after) -
            (allocated_objects_before-live_objects_before));
    
    fprintf(stderr, "Number of Increments: %d\n", number_of_inc);
    fprintf(stderr, "Number of Decrements: %d\n", number_of_dec);
}
