/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include "ThreadedCodeGeneration.h"
#include "Interpreter_threaded.h"

#include <module.VMObjects>
#include <module.Bytecode>
#include <module.VM>

#include <misc/debug.h>

#include <stdbool.h>
#include <stdio.h>

// Look for superinstructions in bytecode and patch threaded code accordingly
size_t tcode_patch_superinstructions(pVMMethod _method, size_t i, size_t num_bytecodes, size_t min_seq_length) {			

    bytecode_t initial_bc = SEND(_method, get_bytecode, i);
    
    // First, determine the max sequence length we can even consider due to how many
    // remainig bytecodes there are in the method code.
	size_t max_sequence_length = SUPER_INST_MAX_SEQUENCE_LENGTH;
	if ((num_bytecodes - i) < max_sequence_length)
		max_sequence_length = num_bytecodes - i;
    
	if ((initial_bc < Interpreter_threaded_first_super_bc()) && (max_sequence_length > 1)) {
        // Now look for the longest sequence matching a super instruction.
		// To do so, compare from current bytecode pos against all sequences of
		// all super instructions, keep the longest.

		size_t  longest_seq = 1;
		uint8_t longest_seq_bc = SUPER_INST_SEQUENCE_SENTINEL; 
		size_t  longest_seq_bc_size = 0;
        
		for (size_t n = Interpreter_threaded_first_super_bc(); n <= Interpreter_threaded_last_super_bc(); n++) {
			// check if this is the current longest sequence
			size_t seq_entry_idx = 0;
			size_t seq_bc_size = 0;
			
            while ((i + seq_bc_size) < num_bytecodes) {
				bytecode_t this_bc = SEND(_method, get_bytecode, i + seq_bc_size);
				bytecode_t seq_bc = inst_handler_entries[n].sequence[seq_entry_idx++];
                
				if (seq_bc == SUPER_INST_SEQUENCE_SENTINEL) {
                    // Reached the end of this sequence as marked by the
                    // sentinel value. Determine if this is the longest
                    // sequence we found so far and update accordingly.
					if (seq_entry_idx > longest_seq) {
						longest_seq = seq_entry_idx;
						longest_seq_bc_size = seq_bc_size;
						longest_seq_bc = n;
						break;
					}
				}
				else {
					if (this_bc == seq_bc) {
                        // Bytecode still in sequence.
						seq_bc_size += bytecodes_get_bytecode_length(this_bc);
					}
					else {
                        // Bytecode broke the sequence.
						break;
					}
				}
			}
		}
        
		if ((longest_seq > min_seq_length) &&
            (longest_seq_bc != SUPER_INST_SEQUENCE_SENTINEL)) {
            // We have a superinstruction for the current sequence.
            // Use it by patching its label into the place of the first threaded code
            // of the sequence.
			SEND(_method, set_bytecode, i, longest_seq_bc);
            SEND(_method, set_tcode, i, inst_handler_entries[longest_seq_bc].label);
            return longest_seq_bc_size;
		}
	}
	
	return bytecodes_get_bytecode_length(initial_bc);
}

pVMMethod tcode_dynamic_convert(pVMMethod _method) {
    printf("tcode_dynamic_convert\n");
	//create a new method that's identical with the parameter,
	//but provides enough space to store the threaded code
	pVMMethod meth = VMMethod_new_with(SEND((pVMInvokable)_method, get_signature), SEND(_method, get_number_of_bytecodes), 
										 SEND(_method, get_number_of_indexable_fields));
	
    for (size_t i = 0; i < SEND(meth, get_number_of_bytecodes); i++) {
		uint8_t cur_bc = SEND(_method, get_bytecode, i);
		SEND(meth, set_bytecode, i, cur_bc);
	}

	return _method;
}

void tcode_dynamic_translate(pVMMethod _method) {
	
	//get number of bytecodes
	size_t num_bytecodes = SEND(_method, get_number_of_bytecodes);

	//if there is not enough space for the tcode, a new method is created
	//this is probably illegal?
	if (SEND(_method, get_number_of_tcodes) < num_bytecodes) {
		_method = tcode_dynamic_convert(_method);
	}
	
    for (size_t i = 0; i < num_bytecodes; i++) {
		//translate bytecode to threaded code label and set it into the tcode
        bytecode_t bc = SEND(_method, get_bytecode, i);
		void* tcode = inst_handler_entries[bc].label;
		SEND(_method, set_tcode, i, tcode);
    
        for (size_t j = 1; j < bytecodes_get_bytecode_length(bc); j++) {
            i++;
            bytecode_t nextbc = SEND(_method, get_bytecode, i);
            SEND(_method, set_tcode, i, (void*)(intptr_t)nextbc);
        }
    }
    
	//iterate through bytecodes
	if (Interpreter_threaded_first_super_bc() != Interpreter_threaded_last_super_bc()) {
        for (size_t n = SUPER_INST_MAX_SEQUENCE_LENGTH; n >= 2; n--) {
            for (size_t i = 0; i < num_bytecodes;) {
                //translate bytecode to threaded code label and set it into the tcode
                void* tcode = inst_handler_entries[SEND(_method, get_bytecode, i)].label;
                SEND(_method, set_tcode, i, tcode);
        
                //Replace with superinstruction if applicable and calculate number of bytecodes  
                //to skip, i.e. copied directly into the threaded code. They are either 
                //arguments or will be skipped anyways as they are part of the superinstruction, 
                //so they can be copied directly into the threaded code.
            
                i += tcode_patch_superinstructions(_method, i, num_bytecodes, n);
            }
        }
    }
}

