#include <stdint.h>

#include <module.Image>

/**
 * \brief Updates all pointers to referenced heap-located objects.
 * 
 * In case of a loaded or relocated heap, all object references in 
 * this particular object are updated. The method is invoked recursively
 * on all these objects.
 * 
 * \param _self The object in question.
 * \param difference The difference by which to change the object pointers.
 */
void _VMObject_update_references(void* _self, const intptr_t difference) {
	pVMObject self = (pVMObject) _self;

	// mark object as processed
	self->gc_field = REFERENCES_UPDATED;
	
	// iterate through fields and update them
	size_t number = SEND(self, get_number_of_fields);
	for (size_t curr_field_idx = 0; curr_field_idx < number; curr_field_idx++) {
		
		pVMObject old = SEND(self, get_field, curr_field_idx);
		pVMObject anew = (pVMObject)((intptr_t)old + difference);
		
		// managed object?
		if (!gc_is_managed_object(anew))
			continue;
		
		SEND(self, set_field, curr_field_idx, anew);
		Image_update_object_references(anew, difference);
	}
}


/**
 * \brief Store the object's non-heap data to the image file.
 * 
 * The object saves its non-heap data, i.e. strings in the image file.
 * The method is invoked recursively for all referenced objects.
 * 
 * \param _self The object in question.
 * \return True in case of successful saving of data, false otherwise.
 * 
 * \note This method may be removed when all objects are fully heap-contained.
 * \note This method is meant to be overridden, as an object of VMObject
 * does not have any data to be serialized.
 */
bool _VMObject_serialize_data(void* _self) {
	pVMObject self = (pVMObject)_self;
	
	self->gc_field = OBJECT_DATA_SERIALIZED;
	
	// iterate through fields and update them
	// skip class field
	int number = SEND(self, get_number_of_fields);
	for (int curr_field_idx = 0; curr_field_idx < number; ++curr_field_idx) {

		if (!Image_serialize_object_data(SEND(self, get_field, curr_field_idx)))
				return false;
	}

	return true;
}


/**
 * \brief Deserializes the object's non-heap data from the image file.
 * 
 * The object loads its non-heap data, i.e. strings from the image file.
 * The method is invoked recursively for all referenced objects.
 * 
 * \param _self The object in question.
 * \return True in case of successful restoration of data, false otherwise.
 * 
 * \note This method may be removed when all objects are fully heap-contained.
 * \note This method is meant to be overridden, as an object of VMObject
 * does not have any data to be serialized.
 */
bool _VMObject_deserialize_data(void* _self) {
	pVMObject self = (pVMObject)_self;
	
	self->gc_field = OBJECT_DATA_RESTORED;

	// iterate through fields and update them
	// skip class field
	int number = SEND(self, get_number_of_fields);
	for (int curr_field_idx = 0; curr_field_idx < number; ++curr_field_idx) {

		if (!Image_deserialize_object_data(SEND(self, get_field, curr_field_idx)))
			return false;
	}

	return true;
}
