/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include "Profiler.h"

#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <math.h>

#include <module.Bytecode>

#include <module.VM>

int** stats;//doubles
int** stats2;//triples
int** stats3;//quads

//file to write the xml profile to
FILE* fp;

//flag if the profiler is initialized
int initialized = 0;

//history
uint8_t lastbc;				//last bytecode
int last_super_index;		//last two bytecodes
int last_super_super_index; //last three bytecodes

/* Resets the profiler. Automatically called on first access to profiler if it
   was not done explicitly before.   */
void Profiler_reset() {
	//reset all stats
	if (stats != NULL || stats2 != NULL || stats3 != NULL || fp != NULL) Profiler_close();

	//get memory for the stats
	stats = (int**) malloc(16*sizeof(int*));
	stats2 = (int**) malloc(256*sizeof(int*));
	stats3 = (int**) malloc(4096*sizeof(int*));
	int i = 0;
	for (; i < 16; i++) {
		stats[i] = (int*) malloc(16*sizeof(int));
		stats2[i] = (int*) malloc(16*sizeof(int));
		stats3[i] = (int*) malloc(16*sizeof(int));
		memset(stats[i], 0, sizeof(int) * 16);
		memset(stats2[i], 0, sizeof(int) * 16);
		memset(stats3[i], 0, sizeof(int) * 16);
	}
	for (; i < 256; i++) {
		stats2[i] = (int*) malloc(16*sizeof(int));
		stats3[i] = (int*) malloc(16*sizeof(int));
		memset(stats2[i], 0, sizeof(int) * 16);
		memset(stats3[i], 0, sizeof(int) * 16);
	}
	for (; i < 4096; i++) {
		stats3[i] = (int*) malloc(16*sizeof(int));
		memset(stats3[i], 0, sizeof(int) * 16);
	}

	//reset "history"
	lastbc = 255;
	last_super_index = 256;
	last_super_super_index = 4096;

	//start xml output
	if((fp=fopen("profile.xml", "wb"))==NULL) {
		 printf("Can't open file %s to write.\n", "profile.xml");
         Universe_exit(1);
	}
	fprintf(fp, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<trace>\n\t<basic_block>\n");

	//done initializing
	initialized = 1;
}

/* Adds a bytecode to the profile. */
void Profiler_add(uint8_t bc){
	if (!initialized) {
		Profiler_reset();
		//initialized = 1;
	}
	//write xml output
	fprintf(fp, "\t\t<instruction bytecode=\"%d\" />\n", bc);
	if(bc == BC_SEND || bc == BC_SUPER_SEND || bc == BC_RETURN_LOCAL || bc == BC_RETURN_NON_LOCAL) {
		fprintf(fp, "\t</basic_block>\n\t<basic_block>\n");
	}

	//add to stats
	if (lastbc != 255) {//this is not the first bc in this basic block
		//so we add to the doubles statistics
		stats[lastbc][bc]++;
		if (last_super_index != 256) {//there have been at least 2 instructions before this one
			//so we can add to the triple stats
			stats2[last_super_index][bc]++;
			if (last_super_super_index != 4096) {//there have been at least 3 instructions before
				//so we can add to the quads stats
				stats3[last_super_super_index][bc]++;
			}
			//mark that there have already been three instructions and remember them
			last_super_super_index = last_super_index*16 + bc;
		}
		//mark that there have already been two instructions and remember them
		last_super_index = lastbc*16 + bc;
	}
	//mark that we have read a bytecode and remember it
	lastbc = bc;

	//if this bytecode ends a basic block, we clear the "history"
	if(lastbc == BC_SEND || lastbc == BC_RETURN_LOCAL || lastbc == BC_RETURN_NON_LOCAL) {
		lastbc = 255;
		last_super_index = 256;
		last_super_super_index = 4096;
	}
}

/* Closes the profile, by closing the xml output file, freeing the allocated
   stats memory and marking the profiler as uninitialized        */
void Profiler_close() {
	if (fp != NULL) {
		fprintf(fp, "\t</basic_block>\n</trace>\n");
		fclose(fp);
	}
	if (stats != NULL){
		for (int i = 0; i < 16; i++) free(stats[i]);;
		free(stats);
	}
	if (stats2 != NULL){
		for (int i = 0; i < 256; i++) free(stats2[i]);;
		free(stats2);
	}
	if (stats3 != NULL){
		for (int i = 0; i < 4096; i++) free(stats3[i]);;
		free(stats3);
	}
	initialized = 0;
}

void Profiler_print_top_n(int top_n, int seq_length, int** _stats) {
	if (!initialized) {
		printf("Profiler not initialized, you will see HALT often now...\n");
		Profiler_reset();
	}

	//get memory to store the winners in
	int** tt;
	tt = (int**) malloc(top_n*sizeof(int*));
	if (tt == NULL) {printf("Error allocating memory\n"); Profiler_close(); return;}
	for (int i = 0; i < top_n; i++) {
		tt[i] = (int*) malloc(2* sizeof(int));
		if (tt[i] == NULL) {printf("Error allocating memory\n"); Profiler_close(); return;}
		memset(tt[i], 0, 2*sizeof(int));
	}

	//the size of _stats in the first dimension
	int stats_size = (int)pow(16, seq_length-1);

	//go through the stats
	for (int i = 0; i < stats_size; i++) {
		for (int j = 0; j < 16; j++) {
			for (int k = 0; k < top_n; k++) {
				//compare the current stats entry with the highscore
				if (_stats[tt[k][0]][tt[k][1]] < _stats[i][j]) {
					//we got a new highscore, so we move the others down
					for (int p = top_n-1; p > k; p--)
						for (int v = 0; v < 2; v++)
							tt[p][v] = tt[p-1][v];
					//add new entry
					tt[k][0] = i;
					tt[k][1] = j;
					break;
				}
			}
		}
	}

	printf("---------top %d combinations with length %d----------\n", top_n, seq_length);
	//go through highscore
	for (int i = 0; i < top_n; i++) {
		printf("%d. Platz (%d)\n",i,  _stats[tt[i][0]][tt[i][1]]);
		//depending on the sequence length there may be arbitrary many values encoded in
		//the first value
		for (int j = seq_length-2; j >= 0; j--) {
			//bytecode values are calculated depending on the current level of indirection.
			//the actual bytecode is calculated by bc = (value/ 16^(lod-2)) % 16
			const char* bc_name = bytecode_names[(tt[i][0]/(int)pow(16,j)) % 16];
			printf("%d: %s", seq_length - (j+1), bc_name);
		}
		printf("%d: %s\n", seq_length, bytecode_names[tt[i][1]]);
		//printf("  \n");
	}
	//printf("done\n");
	for (int i = 0; i < top_n; i++) {
		free(tt[i]);
	}
	free(tt);
}

/*prints top twenty combinations of doubles, triples and quads*/
void Profiler_print_top(int top_n) {
	Profiler_print_top_n(top_n, 2, stats);
	Profiler_print_top_n(top_n, 3, stats2);
	Profiler_print_top_n(top_n, 4, stats3);
}


/* Prints the complete preprocessed profile as a table into a html document */
void Profiler_print_html(char* filename) {
	if(!initialized) {
		Profiler_reset();
	}

	FILE *fp;

	if((fp=fopen(filename, "wb"))==NULL) {
		 printf("Can't open file %s to write.\n", filename);
         Universe_exit(1);
	}

	fprintf(fp ,"<html><head></head><body>\n");
	fprintf(fp ,"<table style=\"border-width:1px;border-style:solid\">\n");
	fprintf(fp ,"\t<tr style=\"border-width:1px;border-style:solid\">\n");
	fprintf(fp ,"\t\t<th style=\"border-width:1px;border-style:solid\"></th>\n");
	for (int i = 0; i < 16; i++)
	{
		fprintf(fp ,"\t\t<th style=\"border-width:1px;border-style:solid\">%s</th>\n", bytecode_names[i]);
	}
	fprintf(fp ,"\t</tr>\n");
	for(int i = 0; i < 16; i++) {
		fprintf(fp ,"\t<tr style=\"border-width:1px;border-style:solid\">\n\t\t<th style=\"border-width:1px;border-style:solid\">%s</th>\n", bytecode_names[i]);
		for (int j = 0; j < 16; j++) {

			fprintf(fp ,"\t\t<td style=\"text-align:center;border-width:1px;border-style:solid\">%d</td>\n", stats[i][j]);
		}
		fprintf(fp ,"\t</tr>\n");
	}
	fprintf(fp ,"</table>\n");

	fprintf(fp ,"<table style=\"border-width:1px;border-style:solid\">\n");
	fprintf(fp ,"\t<tr style=\"border-width:1px;border-style:solid\">\n");
	fprintf(fp ,"\t\t<th style=\"border-width:1px;border-style:solid\"></th>\n");
	for (int i = 0; i < 16; i++)
	{
		fprintf(fp ,"\t\t<th style=\"border-width:1px;border-style:solid\">%s</th>\n", bytecode_names[i]);
	}

	for(int i = 0; i < 255; i++) {
		fprintf(fp ,"\t<tr style=\"border-width:1px;border-style:solid\">\n\t\t<th>%s/%s</th>\n", bytecode_names[i/16], bytecode_names[i%16]);
		for (int j = 0; j < 16; j++) {

			fprintf(fp ,"\t\t<td style=\"text-align:center;border-width:1px;border-style:solid\">%d</td>\n", stats2[i][j]);
		}
		fprintf(fp ,"\t</tr>\n");
	}

	fprintf(fp ,"</table>\n");

	fprintf(fp ,"<table style=\"border-width:1px;border-style:solid\">\n");
	fprintf(fp ,"\t<tr style=\"border-width:1px;border-style:solid\">\n");
	fprintf(fp ,"\t\t<th style=\"border-width:1px;border-style:solid\"></th>\n");
	for (int i = 0; i < 16; i++)
	{
		fprintf(fp ,"\t\t<th style=\"border-width:1px;border-style:solid\">%s</th>\n", bytecode_names[i]);
	}

	for(int i = 0; i < 4096; i++) {
		fprintf(fp ,"\t<tr style=\"border-width:1px;border-style:solid\">\n\t\t<th>%s/%s/%s</th>\n", bytecode_names[i/256], bytecode_names[(i/16)%16],bytecode_names[i%16]);
		for (int j = 0; j < 16; j++) {

			fprintf(fp ,"\t\t<td style=\"text-align:center;border-width:1px;border-style:solid\">%d</td>\n", stats3[i][j]);
		}
		fprintf(fp ,"\t</tr>\n");
	}

	fprintf(fp ,"</table>\n</body></html>\n");

	fclose(fp);
}

void Profiler_print() {

	printf("----------combinations of two-------------------\n");

	for(int i = 0; i < 16; i++) {
		for (int j = 0; j < 16; j++) {

			printf("%d ", stats[i][j]);
		}
		printf("\n");
	}
	printf("---------combinations of three------------------\n");

	for(int i = 0; i < 256; i++) {
		for (int j = 0; j < 16; j++) {

			printf("%d ", stats2[i][j]);
		}
		printf("\n");
	}
	printf("---------combinations of four------------------\n");

	for(int i = 0; i < 4096; i++) {
		for (int j = 0; j < 16; j++) {

			printf("%d ", stats3[i][j]);
		}
		printf("\n");
	}
}
