/*
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include <module.VM>
#include <module.VMObjects>

#include <misc/String.h>

#include "VMArray.h"
#include "VMClass.h"

// externally refenced variables:
pVMObject nil_object;
pVMObject true_object;
pVMObject false_object;

pVMClass object_class;
pVMClass class_class;
pVMClass metaclass_class;

pVMClass nil_class;
pVMClass integer_class;
pVMClass biginteger_class;
pVMClass array_class;
pVMClass method_class;
pVMClass symbol_class;
pVMClass frame_class;
pVMClass primitive_class;
pVMClass string_class;
pVMClass system_class;
pVMClass block_class;
pVMClass double_class;


pVMClass new_metaclass_class(void) {
    // Allocate the metaclass classes
    pVMClass result = VMClass_new();
    SEND((pVMObject)result, set_class, VMClass_new());
    
    // Setup the metaclass hierarchy
    pVMObject mclass = (pVMObject)SEND((pVMObject)result, get_class);
    SEND(mclass, set_class, result);
    
    // Return the freshly allocated metaclass class
    return result;
}


pVMClass new_system_class(void) {
    // Allocate the new system class
    pVMClass system_class = VMClass_new();
    
    // Setup the metaclass hierarchy
    SEND((pVMObject)system_class, set_class, VMClass_new());
    pVMObject mclass = (pVMObject)SEND((pVMObject)system_class, get_class);
    SEND(mclass, set_class, metaclass_class);
    
    // Return the freshly allocated system class
    return system_class;
}


void initialize_system_class(pVMClass system_class,
    pVMClass super_class, const char* name
) {
    pString s_name = String_new(name);
    
    // Initialize the superclass hierarchy
    if(super_class != NULL) {
        SEND(system_class, set_super_class, super_class);
        pVMClass sys_class_class = SEND((pVMObject)system_class, get_class);
        pVMClass super_class_class = SEND((pVMObject) super_class, get_class);
        SEND(sys_class_class, set_super_class, super_class_class);
    } else {
        pVMClass sys_class_class = SEND((pVMObject)system_class, get_class);
        SEND(sys_class_class, set_super_class, class_class);
    }

    pVMClass sys_class_class = SEND((pVMObject)system_class, get_class);

    // Initialize the array of instance fields
    SEND(system_class, set_instance_fields, VMArray_new(0));
    SEND(sys_class_class, set_instance_fields, VMArray_new(0));

    // Initialize the array of instance invokables
    SEND(system_class, set_instance_invokables, VMArray_new(0));
    SEND(sys_class_class, set_instance_invokables, VMArray_new(0));
    
    // Initialize the name of the system class
    SEND(system_class, set_name, Universe_symbol_for(s_name));
    pString class_class_name = String_new_from(s_name);
    SEND(class_class_name, concatChars, " class");
    SEND(sys_class_class, set_name, Universe_symbol_for(class_class_name));
    
    // Insert the system class into the dictionary of globals
    Universe_set_global(SEND(system_class, get_name), (pVMObject)system_class);
}


void VMObjects_initialize() {
	// allocate the nil object
    nil_object = VMObject_new();
    
    // allocate the Metaclass classes
    metaclass_class = new_metaclass_class();
    
    // allocate the rest of the system classes
    object_class    = new_system_class();
    nil_class       = new_system_class();
    class_class     = new_system_class();
    array_class     = new_system_class();
    symbol_class    = new_system_class();
    method_class    = new_system_class();
    integer_class   = new_system_class();
    biginteger_class= new_system_class();
    frame_class     = new_system_class();
    primitive_class = new_system_class();
    string_class    = new_system_class();
    double_class    = new_system_class();
        
    // setup the class reference for the nil object
    SEND(nil_object, set_class, nil_class);
        
    // initialize the system classes.
    initialize_system_class(object_class, 	 NULL, 			"Object");
    initialize_system_class(class_class,	 object_class,	"Class");
    initialize_system_class(metaclass_class, class_class, 	"Metaclass");
    initialize_system_class(nil_class, 		 object_class,	"Nil");
    initialize_system_class(array_class, 	 object_class,	"Array");
    initialize_system_class(method_class, 	 array_class,	"Method");
    initialize_system_class(symbol_class, 	 object_class, 	"Symbol");
    initialize_system_class(integer_class, 	 object_class, 	"Integer");
    initialize_system_class(biginteger_class, object_class, "BigInteger");
    initialize_system_class(frame_class, 	 array_class,	"Frame");
    initialize_system_class(primitive_class, object_class, 	"Primitive");
    initialize_system_class(string_class, 	 object_class, 	"String");
    initialize_system_class(double_class, 	 object_class, 	"Double");
}

pVMClass VMObjects_initialize_class(const char* class_name) {
	pVMClass clazz = new_system_class();
	initialize_system_class(clazz, object_class, class_name);
	return clazz;
}
